import abc
import collections
import collections.abc
import dataclasses
import io
import textwrap
import xml.parsers
import xml.sax
import xml.sax.xmlreader

import typing_extensions as typing

from sila import datetime

from .parse_error import ParseError

T = typing.TypeVar("T")
S = typing.TypeVar("S")


@dataclasses.dataclass
class Token(abc.ABC):
    pass


@dataclasses.dataclass
class StartElement(Token):
    name: str
    attrs: dict[str, str] = dataclasses.field(compare=False, default_factory=dict)


@dataclasses.dataclass
class EndElement(Token):
    name: str


@dataclasses.dataclass
class Characters(Token):
    value: list[str]


@dataclasses.dataclass
class EndDocument(Token):
    pass


class Deserializer(xml.sax.ContentHandler, xml.sax.ErrorHandler, typing.Generic[T]):
    def __init__(self):
        super().__init__()

        self.running = False
        self.content = ""
        self._names: list[str] = []
        self._result: typing.Optional[T] = None
        self._exception: typing.Optional[BaseException] = None
        self._locator: typing.Optional[xml.sax.xmlreader.Locator] = None
        self._tokens: list[Token] = []
        self._characters: list[str] = []
        self._handlers = collections.deque[collections.abc.Generator[None, typing.Any, typing.Any]]()

    @classmethod
    def deserialize(
        cls,
        content: str,
        handler: typing.Callable[["Deserializer"], collections.abc.Generator[None, typing.Any, T]],
    ) -> T:
        """
        Deserialize the given XML string using the given handler.

        Args:
          content: The string representation of the XML data.
          handler: The handler to deserialize the root element.

        Returns:
          The parsed object.

        Raises:
          ParsingError: If unexpected or invalid data is detected during
            parsing.
        """

        deserializer = cls()
        deserializer.register(handler(deserializer))
        deserializer.content = content
        parser = xml.sax.make_parser()
        parser.setContentHandler(deserializer)
        parser.setErrorHandler(deserializer)
        parser.forbid_dtd = True  # type: ignore
        parser.forbid_entities = True  # type: ignore
        parser.forbid_external = True  # type: ignore

        stream = io.StringIO(content.strip())
        parser.parse(stream)

        return deserializer.result()

    def done(self) -> bool:
        """Whether the deserializer has a result or an exception set."""

        return self._result is not None or self._exception is not None

    def result(self) -> T:
        """
        Return the result of the deserializer.

        Returns:
          The result value.

        Raises:
          RuntimeError: If the deserializer's result isn't yet available.
          BaseException: If the deserializer is done and has an exception
            set by the `set_exception()` method.
        """

        if self._exception is not None:
            raise self._exception

        if self._result is None:
            msg = "Result is not set."
            raise RuntimeError(msg)

        return self._result

    def set_result(self, result: T) -> None:
        """
        Mark the deserializer as done and set its result.

        Args:
          result: The result to set.
        """

        self._result = result

    def set_exception(self, exception: BaseException) -> None:
        """
        Mark the deserializer as done and set an exception.

        Args:
          exception: The exception to raise.
        """

        self._exception = exception

    def read_start_element(self, name: str) -> collections.abc.Generator[None, Token, StartElement]:
        """
        Expect the start of an element with the given name.

        Args:
          name: Contains the raw XML 1.0 name of the element type.

        Yields:
          The detected start element.

        Raises:
          ParsingError: If the expected and detected tokens differ.
        """

        if self._tokens:
            token = self._tokens.pop()
        else:
            token = yield

        if not isinstance(token, StartElement):
            if isinstance(token, EndElement):
                msg = f"Expected start element with name '{name}', received end element with name '{token.name}'."
                raise ValueError(msg)

            if isinstance(token, Characters):
                msg = f"Expected start element with name '{name}', received characters '{token.value}'."
                raise ValueError(msg)

            msg = f"Expected start element with name '{name}', received token '{token}'."
            raise ValueError(msg)

        if token.name != name:
            msg = f"Expected start element with name '{name}', received start element with name '{token.name}'."
            raise ValueError(msg)

        return token

    def read_end_element(self, name: str) -> collections.abc.Generator[None, Token, EndElement]:
        """
        Expect the end of an element with the given name.

        Args:
          name: Contains the raw XML 1.0 name of the element type.

        Yields:
          The detected end element.

        Raises:
          ParsingError: If the expected and detected tokens differ.
        """

        if self._tokens:
            token = self._tokens.pop()
        else:
            token = yield

        if not isinstance(token, EndElement):
            if isinstance(token, StartElement):
                msg = f"Expected end element with name '{name}', received start element with name '{token.name}'."
                raise ValueError(msg)

            if isinstance(token, Characters):
                msg = f"Expected end element with name '{name}', received characters '{token.value}'."
                raise ValueError(msg)

            msg = f"Expected end element with name '{name}', received token '{token}'."
            raise ValueError(msg)

        if token.name != name:
            msg = f"Expected end element with name '{name}', received end element with name '{token.name}'."
            raise ValueError(msg)

        return token

    def read_str(self) -> collections.abc.Generator[None, Token, str]:
        """
        Read a string value from the buffer.

        Returns:
          The string value read from the xml.
        """

        token = yield from self.read_characters()

        return textwrap.dedent("\n".join(token.value)).replace("\n", " ")

    def read_bool(self) -> collections.abc.Generator[None, Token, bool]:
        """
        Read a boolean value from the buffer.

        Returns:
          The boolean value read from the xml.
        """

        value = yield from self.read_str()

        if value not in ("Yes", "No"):
            msg = f"Could not convert '{self._names[-1]}' with value '{value}' to boolean."
            raise ValueError(msg)

        return value == "Yes"

    def read_int(self) -> collections.abc.Generator[None, Token, int]:
        """
        Read an integer value from the buffer.

        Returns:
          The integer value read from the xml.
        """

        value = yield from self.read_str()

        try:
            return int(float(value))
        except ValueError:
            msg = f"Could not convert '{self._names[-1]}' with value '{value}' to integer."
            raise ValueError(msg) from None

    def read_float(self) -> collections.abc.Generator[None, Token, float]:
        """
        Read a float value from the buffer.

        Returns:
          The float value read from the xml.
        """

        value = yield from self.read_str()

        try:
            return float(value)
        except ValueError:
            msg = f"Could not convert '{self._names[-1]}' with value '{value}' to float."
            raise ValueError(msg) from None

    def read_date(self) -> collections.abc.Generator[None, Token, datetime.date]:
        """
        Read a date value from the buffer.

        Returns:
          The date value read from the xml.
        """

        value = yield from self.read_str()

        try:
            date = datetime.date.fromisoformat(value[:10])

            if len(value) > 10:
                if value[10:] == "Z":
                    date = date.replace(tzinfo=datetime.timezone.utc)
                else:
                    tzsign = -1 if value[10] == "-" else 1
                    offset = datetime.timedelta(hours=int(value[11:13]), minutes=int(value[14:16]))
                    date = date.replace(tzinfo=datetime.timezone(tzsign * offset))

            return date
        except ValueError:
            msg = f"Could not convert '{self._names[-1]}' with value '{value}' to date."
            raise ValueError(msg) from None

    def read_time(self) -> collections.abc.Generator[None, Token, datetime.time]:
        """
        Read a time value from the buffer.

        Returns:
          The time value read from the xml.
        """

        value = yield from self.read_str()

        try:
            return datetime.time.fromisoformat(value.replace("Z", "+00:00"))
        except ValueError:
            msg = f"Could not convert '{self._names[-1]}' with value '{value}' to time."
            raise ValueError(msg) from None

    def read_datetime(self) -> collections.abc.Generator[None, Token, datetime.datetime]:
        """
        Read a datetime value from the buffer.

        Returns:
          The datetime value read from the xml.
        """

        value = yield from self.read_str()

        try:
            return datetime.datetime.fromisoformat(value.replace("Z", "+00:00"))
        except ValueError:
            msg = f"Could not convert '{self._names[-1]}' with value '{value}' to datetime."
            raise ValueError(msg) from None

    def read_characters(self) -> collections.abc.Generator[None, Token, Characters]:
        """
        Expect chunks of characters.

        Yields:
          The detected chunks of characters.

        Raises:
          ParsingError: If the expected and detected tokens differ.
        """

        if self._tokens:
            token = self._tokens.pop()
        else:
            token = yield

        if not isinstance(token, Characters):
            if isinstance(token, StartElement):
                msg = f"Expected characters, received start element with name '{token.name}'."
                raise ValueError(msg)
            if isinstance(token, EndElement):
                self._tokens.append(token)
                return Characters(value=[])

            msg = f"Expected characters, received token '{token}'."
            raise ValueError(msg)

        return token

    def peek(self) -> collections.abc.Generator[None, typing.Any, Token]:
        """
        Look ahead one item without advancing the iterator.

        Returns:
          The token that will be next returned from `read()`.
        """

        token = yield

        self._tokens.append(token)

        return token

    def read(
        self, generator: typing.Optional[collections.abc.Generator[None, typing.Any, S]] = None
    ) -> collections.abc.Generator[None, typing.Any, S]:
        """
        Expect to read elements in the order defined in the given handler.

        Args:
          generator: The handler used to continue deserialization.

        Returns:
          The deserialized object returned by the handler.

        Raises:
          ParsingError: If the expected and detected tokens differ.
        """

        if generator is not None:
            self.register(generator)

        token = yield

        return token

    @typing.override
    def startDocument(self) -> None:
        """Signals the beginning of a document."""

        self.running = True

    @typing.override
    def endDocument(self) -> None:
        """Signals the end of a document."""

        self.running = False
        while self._handlers:
            self.__handle_token(EndDocument())

    @typing.override
    def startElement(self, name: str, attrs: typing.Optional[xml.sax.xmlreader.AttributesImpl] = None) -> None:
        """
        Signals the start of an element in non-namespace mode.

        Args:
          name: Contains the raw XML 1.0 name of the element type.
          attrs: Contains the attributes of the element.
        """

        self._names.append(name)

        if not self._handlers:
            msg = f"Received start element with name '{name}', but no handler registered."
            raise ValueError(msg)

        self.__handle_token(StartElement(name, dict(attrs.items()) if attrs is not None else {}))

    @typing.override
    def endElement(self, name: str) -> None:
        """
        Signals the end of an element in non-namespace mode.

        Args:
          name: Contains the raw XML 1.0 name of the element type.
        """

        if not self._handlers:
            msg = f"Received end element with name '{name}', but no handler registered."
            raise ValueError(msg)

        if not self._names:
            msg = f"Did not expect an end element, received end element with name '{name}'."
            raise ValueError(msg)

        if self._names[-1] != name:
            msg = f"Expected end element with name '{self._names[-1]}', received end element with name '{name}'."
            raise ValueError(msg)

        self.__handle_token(EndElement(name))
        self._names.pop()

    @typing.override
    def characters(self, content: str) -> None:
        """
        Receive notification of character data.

        Args:
          content: A chunk of character data.
        """

        if content.strip():
            self._characters.append(content)

    @typing.override
    def setDocumentLocator(self, locator: xml.sax.xmlreader.Locator) -> None:
        """
        Receive access to a locator of document events.

        Args:
          locator: Allows for locating the origin of document events.
        """

        self._locator = locator

    @typing.override
    def error(self, exception: BaseException) -> typing.NoReturn:
        """
        Handle a recoverable error.

        Args:
          exception: The error to handle.
        """

        if not isinstance(exception, ParseError):
            exception = ParseError(
                exception.args[0],
                path=self._names,
                line=self._locator.getLineNumber() or 0 if self._locator is not None else 0,
                column=self._locator.getColumnNumber() or 0 if self._locator is not None else 0,
            )

        self.set_exception(exception)

        raise exception

    @typing.override
    def fatalError(self, exception: BaseException) -> typing.NoReturn:
        """
        Handle a non-recoverable error.

        Args:
          exception: The fatal error to handle.
        """

        self.error(exception)

    def register(self, handler: collections.abc.Generator[None, typing.Any, typing.Any]) -> None:
        """
        Register a handler and advance it to its first yield.

        Args:
          handler: A generator that receives and processes token through
            yield statements.
        """

        self._handlers.append(handler)

        try:
            next(handler)
        except Exception as error:  # noqa: BLE001
            self.error(error)

    def __handle_token(self, token: Token) -> None:
        if self._characters:
            characters = Characters(self._characters)
            self._characters = []

            self.__handle_token(characters)

        try:
            self._handlers[-1].send(token)
        except StopIteration as result:
            self._handlers.pop().close()

            if self._handlers:
                self.__handle_token(result.value)
            else:
                self.set_result(result.value)
        except Exception as error:  # noqa: BLE001
            self.error(error)



from .any import Any
from .binary import Binary
from .boolean import Boolean
from .constrained import Constrained
from .convertable import Convertable, Native
from .custom import Custom
from .data_type import BasicType, DataType
from .date import Date
from .duration import Duration
from .element import Element
from .integer import Integer
from .list import List
from .real import Real
from .string import String
from .structure import Structure
from .time import Time
from .timestamp import Timestamp
from .timezone import Timezone
from .void import Void

__all__ = [
    "Any",
    "BasicType",
    "Binary",
    "Boolean",
    "Constrained",
    "Convertable",
    "Custom",
    "DataType",
    "Date",
    "Duration",
    "Element",
    "Integer",
    "List",
    "Native",
    "Real",
    "String",
    "Structure",
    "Time",
    "Timestamp",
    "Timezone",
    "Void",
]



import grpc
import typing_extensions as typing

if typing.TYPE_CHECKING:
    from ..common import Context
    from ..data_types import Element, Structure


class Protobuf:
    """
    A collection of protobuf messages and services.
    """

    def __init__(self, context: "Context") -> None:
        self.__context = context
        self.__messages: dict[str, type["Structure"]] = {}
        self.__services: dict[str, dict[str, grpc.RpcMethodHandler]] = {}

    @property
    def messages(self) -> dict[str, type["Structure"]]:
        """
        Receive all protobuf message registered in this collection.
        """

        return self.__messages

    @property
    def services(self) -> dict[str, dict[str, grpc.RpcMethodHandler]]:
        """
        Receive all protobuf services registered in this collection.
        """

        return self.__services

    def get_message(self, path: str, /) -> type["Structure"]:
        """
        Receive a protobuf message from the collection of registered
        messages.

        Args:
          path: The complete path of the registered message including its
            package and name.

        Returns:
          The registered protobuf message.

        Raises:
          ValueError: If no message is registered under the given path.
        """

        try:
            return self.__messages[path]
        except KeyError:
            raise ValueError(f"Could not find any message registered under the given path '{path}'.")

    @typing.overload
    def register_message(
        self, /, *, message: type["Structure"], package: typing.Optional[str] = None
    ) -> type["Structure"]:
        """
        Register a protobuf message in the optional package. Overrides
        previously registered messages with the same package and name
        combination.

        Args:
          message: The protobuf message to register.
          package: The package name to register the message under.

        Returns:
          The registered protobuf message.
        """

    @typing.overload
    def register_message(
        self,
        /,
        *,
        name: str,
        message: dict[str, "Element"],
        description: typing.Optional[str] = None,
        package: typing.Optional[str] = None,
    ) -> type["Structure"]:
        """
        Register a protobuf message with the given name in the optional
        package. Overrides previously registered messages with the same
        package and name combination.

        Args:
          name: The name of the protobuf message to register.
          message: The protobuf message to register.
          description: An optional description of the protobuf message.
          package: The package name to register the message under.

        Returns:
          The registered protobuf message.
        """

    def register_message(
        self,
        /,
        *,
        message: typing.Union[type["Structure"], dict[str, "Element"]],
        name: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        package: typing.Optional[str] = None,
    ) -> type["Structure"]:
        from ..data_types import Structure

        if not isinstance(message, dict):
            name = message.__name__

        assert name
        if isinstance(message, dict):
            message = Structure.create(name=name, description=description, elements=message)

        path = f"{package}.{name}" if package is not None else name
        self.__messages[path] = message

        return message

    def register_service(
        self,
        /,
        name: str,
        service: dict[str, grpc.RpcMethodHandler],
        *,
        package: typing.Optional[str] = None,
        force: bool = False,
    ) -> dict[str, grpc.RpcMethodHandler]:
        """
        Register a protobuf service with the given name in the optional
        package. If a service with the same package and name combination
        already exists, the methods of both services are merged and the
        merged service is registered. Methods with the same name override
        previously registered methods. Use the `force` flag to completely
        replace existing services.

        Args:
          name: The name of the protobuf service to register.
          service: The protobuf service to register containing the
            available methods.
          package: The package name to register the service under.
          force: Set True to completely replace any existing service with
            the same package and name combination and prevent merging.

        Returns:
          The registered protobuf service with its methods.
        """

        path = f"{package}.{name}" if package is not None else name

        if not force:
            service = self.__services.get(path, {}) | service

        self.__services[path] = service

        return service

    async def decode(self, path: str, buffer: bytes) -> dict[str, typing.Any]:
        """
        Use a registered message to decode a byte stream into a python
        dictionary.

        Args:
          path: The path of the registered message used for decoding.
          buffer: The byte stream to decode.

        Returns:
          The decoded message as a dictionary.

        Raises:
          ValueError: If no message is registered under the given path.
          DecodeError: If there was an error while decoding the message.
          ConversionError: If there was an error while converting the
            message to its native python format.
        """

        data_type = self.get_message(path)
        data_value = data_type.decode(buffer)
        value = await data_value.to_native(self.__context)

        return value

    async def encode(self, path: str, value: dict[str, typing.Any]) -> bytes:
        """
        Use a registered message to encode a python dictionary into a
        byte stream.

        Args:
          path: The path of the registered message used for encoding.
          value: The python dictionary to encode.

        Returns:
          The encoded message as a byte stream.

        Raises:
          ValueError: If no message is registered under the given path.
          EncodeError: If there was an error while encoding the message.
          ConversionError: If there was an error while converting the
            message from its native python format.
        """

        data_type = self.get_message(path)
        data_value = await data_type.from_native(self.__context, value)
        buffer = data_value.encode()

        return buffer

    def merge(self, protobuf: "Protobuf") -> None:
        """
        Merge another collection of protobuf messages and services in
        place.

        Args:
          protobuf: The other protobuf messages and services to merge.
        """

        self.__messages.update(protobuf.__messages)
        for path, service in protobuf.__services.items():
            self.__services[path] = self.__services.get(path, {}) | service



import collections.abc
import dataclasses

import typing_extensions as typing

from ..data_types.list import List
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType


@dataclasses.dataclass
class ElementCount(Constraint[list]):
    """
    A constraint that enforces an exact element count for a list.

    This constraint ensures that a given list contains exactly the
    specified number of elements. It raises errors if the input is
    not a list or if the list contains a number of elements different
    from the expected count.

    Args:
      value: The exact number of elements the list must contain. Must
        be a non-negative integer and less than 2⁶³.

    Raises:
      ValueError: If `value` is negative or exceeds 2⁶³.
    """

    value: int
    """
    The exact number of elements the list must contain. Must be a
    non-negative integer and less than 2⁶³.
    """

    def __post_init__(self):
        if self.value < 0:
            raise ValueError(f"Element count must be a non-negative integer, received '{self.value}'.")
        if self.value >= 2**63:
            raise ValueError(f"Element count must be less than 2⁶³, received '{self.value}'.")

    async def validate(self, value: list) -> bool:
        if not isinstance(value, list):
            raise TypeError(f"Expected value of type 'list', received '{type(value).__name__}'.")

        if len(value) != self.value:
            raise ValueError(f"Expected list with element count '{self.value}', received '{value}'.")

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.write_int("ElementCount", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, List):
            raise ValueError(f"Expected constraint's data type to be 'List', received '{data_type.__name__}'.")

        yield from deserializer.read_start_element(name="ElementCount")
        value = yield from deserializer.read_int()
        yield from deserializer.read_end_element(name="ElementCount")

        return cls(value)



import dataclasses
import inspect
import logging

import grpc
import grpc.aio
import typing_extensions as typing

from .. import framework
from ..framework import (
    DefinedExecutionError,
    Element,
    Feature,
    MetadataIdentifier,
    Native,
    NoMetadataAllowed,
    Server,
    SiLAError,
    UndefinedExecutionError,
)


@dataclasses.dataclass
class UnobservableProperty(framework.UnobservableProperty):
    """A property describes certain aspects of a SiLA server that do not require an action on the SiLA server."""

    function: typing.Callable = dataclasses.field(repr=False, default=lambda **_: ...)
    """The implementation which is executed by the RPC handler."""

    @property
    def logger(self) -> logging.Logger:
        """A python logger instance."""

        return logging.getLogger(__name__)

    @typing.override
    async def read(self, metadata: typing.Optional[dict[str, bytes]] = None) -> bytes:
        assert self.feature is not None and isinstance(self.feature.context, Server)

        try:
            # Metadata
            if (
                metadata
                and self.fully_qualified_identifier.feature_identifier == "org.silastandard/core/SiLAService/v1"
            ):
                msg = "No metadata allowed for the SiLA Service feature."
                raise NoMetadataAllowed(msg)

            parsed_metadata = dict[MetadataIdentifier, Native]()
            for interceptor in self.feature.context.get_metadata_by_affect(self.fully_qualified_identifier):
                metadatum = await interceptor.from_buffer(self, metadata)
                parsed_metadata.update(await metadatum.intercept(self))

            # Execute
            response = self.function(metadata=parsed_metadata)

            if inspect.isawaitable(response):
                response = await response

            return await self.feature.context.protobuf.encode(
                f"{self.feature.rpc_package}.Get_{self.identifier}_Responses", response
            )
        except SiLAError as error:
            if isinstance(error, DefinedExecutionError) and error._identifier is None:
                error = error.with_feature(self.feature.fully_qualified_identifier)

            raise error
        except Exception as error:
            self.logger.exception(error)
            raise UndefinedExecutionError(str(error)) from error

    async def read_rpc_handler(self, request: bytes, context: grpc.aio.ServicerContext) -> bytes:
        """
        Handle the gRPC call to read the current value of the property.

        Args:
          request: The request payload in protobuf ecoding.
          context: The gRPC call context.

        Returns:
          The response payload in protobuf ecoding.
        """

        try:
            metadata: dict[str, bytes] = {
                key: value
                for key, value in context.invocation_metadata() or ()
                if key.startswith("sila-") and key.endswith("-bin") and isinstance(value, bytes)
            }

            return await self.read(metadata)
        except SiLAError as error:
            raise await error.to_rpc_error(context) from None

    @typing.override
    def add_to_feature(self, feature: "Feature") -> typing.Self:
        super().add_to_feature(feature)

        feature.context.protobuf.register_message(
            name=f"Get_{self.identifier}_Responses",
            message={
                self.identifier: Element(
                    identifier=self.identifier,
                    display_name=self.display_name,
                    description=self.description,
                    data_type=self.data_type,
                )
            },
            package=feature.rpc_package,
        )
        feature.context.protobuf.register_service(
            feature.identifier,
            {f"Get_{self.identifier}": grpc.unary_unary_rpc_method_handler(self.read_rpc_handler)},
            package=feature.rpc_package,
        )

        return self



from .display_name import check_display_name
from .identifier import check_identifier
from .uuid import check_uuid
from .version import VersionLevel, check_version

__all__ = ["VersionLevel", "check_display_name", "check_identifier", "check_uuid", "check_version"]



from .connection_error import SiLAConnectionError
from .defined_execution_error import DefinedExecutionError
from .framework_error import (
    CommandExecutionNotAccepted,
    CommandExecutionNotFinished,
    FrameworkError,
    FrameworkErrorType,
    InvalidCommandExecutionUUID,
    InvalidMetadata,
    NoMetadataAllowed,
)
from .sila_error import SiLAError
from .undefined_execution_error import UndefinedExecutionError
from .validation_error import ValidationError

__all__ = [
    "CommandExecutionNotAccepted",
    "CommandExecutionNotFinished",
    "DefinedExecutionError",
    "FrameworkError",
    "FrameworkError",
    "FrameworkErrorType",
    "FrameworkErrorType",
    "InvalidCommandExecutionUUID",
    "InvalidMetadata",
    "NoMetadataAllowed",
    "SiLAConnectionError",
    "SiLAError",
    "UndefinedExecutionError",
    "ValidationError",
]



import collections.abc
import dataclasses
import re

import typing_extensions as typing

from ..data_types.string import String
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List


@dataclasses.dataclass
class Pattern(Constraint[str]):
    """
    A constraint that enforces a specific regular expression pattern
    for a string value.

    This constraint ensures that a given string matches a specified
    regular expression pattern. It raises errors if the input value
    is not a string or if the string does not match the defined
    pattern.

    Args:
      value: The regular expression pattern that the string value
        must match.
    """

    value: str
    """
    The regular expression pattern that the string value must
    match.
    """

    async def validate(self, value: str) -> bool:
        if not isinstance(value, str):
            raise TypeError(f"Expected value of type 'str', received '{type(value).__name__}'.")

        if not re.fullmatch(self.value, value):
            raise ValueError(f"Value '{value}' does not match the pattern: '{self.value}'.")

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.write_str("Pattern", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, String):
            raise ValueError(f"Expected constraint's data type to be 'String', received '{data_type.__name__}'.")

        yield from deserializer.read_start_element(name="Pattern")
        value = yield from deserializer.read_str()
        yield from deserializer.read_end_element(name="Pattern")

        return cls(value)



import typing_extensions as typing

if typing.TYPE_CHECKING:
    from .observable_command import ObservableCommand


class CommandExecution(typing.Protocol):
    """Tracks the progress of an observable command execution."""

    command_execution_uuid: str
    """Uniquely identifies the Command execution."""

    command: "ObservableCommand"
    """The observable command being executed."""

    def cancel(self) -> None:
        """Cancel the ongoing command execution."""



import dataclasses
import datetime
import time
import uuid

import typing_extensions as typing


@dataclasses.dataclass
class BinaryTransfer:
    """
    Storage and retrieval of chunks for binary data transfers.

    Attributes:
      max_chunk_size: The maximum size of each chunk in bytes.
      binary_transfer_uuid: A unique identifier for the binary
        transfer.
      created_at: The timestamp when the binary transfer was created.
      valid_for: Fractional seconds for how long the binary transfer
        is valid.
      size: The total size of the binary data in bytes.
      chunks: A list of chunks containing the binary data.
    """

    max_chunk_size: typing.ClassVar[int] = 2**21

    binary_transfer_uuid: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    created_at: float = dataclasses.field(default_factory=time.monotonic)
    valid_for: float = dataclasses.field(default=60)
    size: int = 0
    chunks: list[typing.Optional[bytes]] = dataclasses.field(default_factory=list)

    @classmethod
    def new(cls, size: int, chunks: int) -> typing.Self:
        """
        Reserve space for a new binary transfer.

        Args:
          size: The amount of bytes to reserve.
          chunks: The number of expected chunks.

        Returns:
          An instance of a new binary transfer.
        """

        return cls(size=size, chunks=[None] * chunks)

    @classmethod
    def from_buffer(cls, buffer: bytes) -> typing.Self:
        """
        Create a new binary transfer with the given buffer.

        Args:
          buffer: The actual bytes data for this binary transfer.

        Returns:
          An instance of a new binary transfer.
        """

        size = len(buffer)

        return cls(
            size=size,
            chunks=[buffer[i : i + cls.max_chunk_size] for i in range(0, size, cls.max_chunk_size)],
        )

    @property
    def is_complete(self) -> bool:
        """Whether the binary transfer has been completed or not."""

        return not any(chunk is None for chunk in self.chunks)

    @property
    def buffer(self) -> bytes:
        """The complete buffer of the binary transfer."""

        return b"".join(chunk or b"" for chunk in self.chunks)

    @property
    def lifetime(self) -> datetime.timedelta:
        """The duration in which this binary transfer is valid."""

        return datetime.timedelta(seconds=self.created_at + self.valid_for - time.monotonic())

    def get_chunk(self, offset: int, length: int) -> bytes:
        """
        Retrieve a specific chunk of binary data.

        Args:
          offset: The starting byte position within the binary data from
            which to begin retrieving the chunk.
          length: The number of bytes to retrieve starting from the
            offset position.

        Returns:
          The requested chunk of binary data.
        """

        if length > self.max_chunk_size:
            msg = (
                f"Expected length of chunk with offset '{offset}' to not exceed the "
                f"maximum size of 2 MiB, received {length} bytes."
            )
            raise ValueError(msg)

        if offset > self.size:
            msg = f"Expected offset to not exceed the binary's size of {self.size} bytes, received {offset} bytes."
            raise ValueError(msg)

        if offset + length > self.size:
            msg = (
                f"Expected length of chunk with offset '{offset}' to not exceed the "
                f"binary's size of {self.size} bytes, received {length} bytes."
            )
            raise ValueError(msg)

        return self.buffer[offset : offset + length]

    def set_chunk(self, index: int, payload: bytes) -> None:
        """
        Set a specific chunk of binary data.

        Args:
          index: The index of the chunk within the binary transfer.
          payload: The actual binary data chunk being set.
        """

        if self.is_complete:
            msg = f"Received chunk with index '{index}' for already completed binary transfer."
            raise ValueError(msg)

        if len(payload) > self.max_chunk_size:
            msg = f"Expected chunk '{index}' to not exceed the maximum size of 2 MiB, received {len(payload)} bytes."
            raise ValueError(msg)

        if index >= len(self.chunks):
            msg = f"Expected chunks up to index '{len(self.chunks) - 1}', received '{index}'."
            raise ValueError(msg)

        if self.chunks[index] is not None:
            msg = f"Received chunk with index '{index}' for already received chunk."
            raise ValueError(msg)

        self.chunks[index] = payload

        current_size = sum(len(chunk) for chunk in self.chunks if chunk is not None)
        if current_size > self.size:
            msg = (
                f"Expected a total size of {self.size} bytes, received "
                f"already {current_size} bytes with chunk '{index}'."
            )
            raise ValueError(msg)



import collections.abc
import dataclasses
import inspect
import logging

import grpc.aio
import typing_extensions as typing

from .. import framework
from ..framework import (
    CommandConfirmation,
    CommandExecutionInfo,
    CommandExecutionUUID,
    ConversionError,
    DecodeError,
    DefinedExecutionError,
    Feature,
    MetadataIdentifier,
    Native,
    Server,
    SiLAError,
    UndefinedExecutionError,
    ValidationError,
)
from .command_execution import CommandExecution


@dataclasses.dataclass
class ObservableCommand(framework.ObservableCommand):
    """Any command for which observing the progress of execution is possible or does make sense."""

    function: typing.Callable = dataclasses.field(repr=False, default=lambda **_: ...)
    """The implementation which is executed by the RPC handler."""

    @property
    def logger(self) -> logging.Logger:
        """A python logger instance."""

        return logging.getLogger(__name__)

    @typing.override
    async def initiate(
        self, request: bytes = b"", metadata: typing.Optional[dict[str, bytes]] = None
    ) -> "CommandConfirmation":
        assert self.feature is not None and isinstance(self.feature.context, Server)

        try:
            # Metadata
            parsed_metadata = dict[MetadataIdentifier, Native]()
            for interceptor in self.feature.context.get_metadata_by_affect(self.fully_qualified_identifier):
                metadatum = await interceptor.from_buffer(self, metadata)
                parsed_metadata.update(await metadatum.intercept(self))

            # Parameters
            try:
                parameters = await self.feature.context.protobuf.decode(
                    f"{self.feature.rpc_package}.{self.identifier}_Parameters", request
                )
            except DecodeError as error:
                raise ValidationError(
                    parameter=f"{self.fully_qualified_identifier}/Parameter/{error.path[0]}", message=error.message
                ) from error
            except ConversionError as error:
                raise ValidationError(
                    parameter=f"{self.fully_qualified_identifier}/Parameter/{error.path[0]}", message=error.message
                ) from error

            # Execute
            command_execution = CommandExecution(command=self, parameters=parameters, metadata=parsed_metadata)
            self.feature.context.add_command_execution(command_execution)
            command_execution.execute()

            return command_execution.command_confirmation
        except DefinedExecutionError as error:
            error = error.with_feature(self.feature.fully_qualified_identifier)
            raise error

    async def initiate_rpc_handler(self, request: bytes, context: grpc.aio.ServicerContext) -> bytes:
        """
        Handle the gRPC call to initiate the execution of the command.

        Args:
          request: The request payload in protobuf ecoding.
          context: The gRPC call context.

        Returns:
          Confirmation that the command execution has been accepted.
        """

        try:
            metadata: dict[str, bytes] = {
                key: value
                for key, value in context.invocation_metadata() or ()
                if key.startswith("sila-") and key.endswith("-bin") and isinstance(value, bytes)
            }

            command_confirmation = await self.initiate(request, metadata)
            return command_confirmation.encode()
        except SiLAError as error:
            raise await error.to_rpc_error(context) from None

    @typing.override
    async def subscribe_status(
        self, command_execution_uuid: str
    ) -> collections.abc.AsyncIterator["CommandExecutionInfo"]:
        assert self.feature is not None and isinstance(self.feature.context, Server)

        command_execution: CommandExecution = self.feature.context.get_command_execution(command_execution_uuid)

        async for status in command_execution.execution_info:
            yield status

    async def subscribe_status_rpc_handler(
        self, request: bytes, context: grpc.aio.ServicerContext
    ) -> collections.abc.AsyncIterator[bytes]:
        """
        Handle the gRPC call to subscribe to status changes of the command execution.

        Args:
          request: The request payload in protobuf ecoding.
          context: The gRPC call context.

        Returns:
          The current status of the command execution.
        """

        try:
            command_execution_uuid = CommandExecutionUUID.decode(request).value
            async for value in self.subscribe_status(command_execution_uuid):
                yield value.encode()
        except SiLAError as error:
            raise await error.to_rpc_error(context) from None

    @typing.override
    async def subscribe_intermediate(self, command_execution_uuid: str) -> collections.abc.AsyncIterator[bytes]:
        assert self.feature is not None and isinstance(self.feature.context, Server)

        command_execution: CommandExecution = self.feature.context.get_command_execution(command_execution_uuid)

        async for intermediate_responses in command_execution.intermediate_responses:
            yield await self.feature.context.protobuf.encode(
                f"{self.feature.rpc_package}.{self.identifier}_IntermediateResponses", intermediate_responses
            )

    async def subscribe_intermediate_rpc_handler(
        self, request: bytes, context: grpc.aio.ServicerContext
    ) -> collections.abc.AsyncIterator[bytes]:
        """
        Handle the gRPC call to subscribe to intermediate responses of the command execution.

        Args:
          request: The request payload in protobuf ecoding.
          context: The gRPC call context.

        Returns:
          The current intermediate responses of the command execution.
        """

        try:
            command_execution_uuid = CommandExecutionUUID.decode(request).value
            async for value in self.subscribe_intermediate(command_execution_uuid):
                yield value
        except SiLAError as error:
            raise await error.to_rpc_error(context) from None

    @typing.override
    async def get_result(self, command_execution_uuid: str) -> bytes:
        assert self.feature is not None and isinstance(self.feature.context, Server)

        command_execution: CommandExecution = self.feature.context.get_command_execution(command_execution_uuid)
        return command_execution.result()

    async def get_result_rpc_handler(self, request: bytes, context: grpc.aio.ServicerContext) -> bytes:
        """
        Handle the gRPC call to get the responses of the command execution.

        Args:
          request: The request payload in protobuf ecoding.
          context: The gRPC call context.

        Returns:
          The resulting responses of the command execution.
        """

        try:
            command_execution_uuid = CommandExecutionUUID.decode(request).value
            return await self.get_result(command_execution_uuid)
        except SiLAError as error:
            raise await error.to_rpc_error(context) from None

    async def execute(
        self,
        parameters: dict[str, Native],
        metadata: dict[MetadataIdentifier, Native],
        command_execution: CommandExecution,
    ) -> None:
        """
        Execute the command.

        Args:
          parameters: Input parameters passed to the command.
          metadata: Additional metadata sent from client to server.
          command_execution: The command execution instance.
        """

        assert self.feature is not None and isinstance(self.feature.context, Server)

        try:
            responses = self.function(**parameters, metadata=metadata, command_execution=command_execution)

            if inspect.isawaitable(responses):
                responses = await responses

            message = await self.feature.context.protobuf.encode(
                f"{self.feature.rpc_package}.{self.identifier}_Responses", responses
            )
            await command_execution.set_result(message)
        except SiLAError as error:
            if isinstance(error, DefinedExecutionError) and error._identifier is None:
                error = error.with_feature(self.feature.fully_qualified_identifier)

            await command_execution.set_exception(error)
        except Exception as error:
            self.logger.exception(error)
            await command_execution.set_exception(UndefinedExecutionError(str(error)))

    @typing.override
    def add_to_feature(self, feature: "Feature") -> typing.Self:
        super().add_to_feature(feature)

        feature.context.protobuf.register_message(
            name=self.identifier + "_Parameters", message=self.parameters, package=feature.rpc_package
        )
        feature.context.protobuf.register_message(
            name=self.identifier + "_Responses", message=self.responses, package=feature.rpc_package
        )
        if self.intermediate_responses:
            feature.context.protobuf.register_message(
                name=self.identifier + "_IntermediateResponses",
                message=self.intermediate_responses,
                package=feature.rpc_package,
            )

        services = {
            self.identifier: grpc.unary_unary_rpc_method_handler(self.initiate_rpc_handler),
            f"{self.identifier}_Info": grpc.unary_stream_rpc_method_handler(self.subscribe_status_rpc_handler),
            f"{self.identifier}_Result": grpc.unary_unary_rpc_method_handler(self.get_result_rpc_handler),
        }
        if self.intermediate_responses:
            services[f"{self.identifier}_Intermediate"] = grpc.unary_stream_rpc_method_handler(
                self.subscribe_intermediate_rpc_handler
            )

        feature.context.protobuf.register_service(feature.identifier, services, package=feature.rpc_package)

        return self



import dataclasses

import typing_extensions as typing

from sila import datetime

from ..protobuf import Message, Reader, WireType, Writer
from .convertable import Convertable

if typing.TYPE_CHECKING:
    from ..common import Context


@dataclasses.dataclass
class Timezone(Message, Convertable):
    """
    A signed, fixed-length span of time representing an offset from UTC.

    Attributes:
      hours: The hours of the timezone value in range [-12-14].
        Defaults to zero.
      minutes: The minutes of the timezone value in range [0-59].
        Defaults to zero.
    """

    hours: int = 0
    minutes: int = 0

    @typing.overload
    @classmethod
    async def from_native(
        cls, context: "Context", value: typing.Optional[datetime.timezone] = None, /
    ) -> typing.Self: ...

    @typing.overload
    @classmethod
    async def from_native(cls, context: "Context", /, *, hours: int = 0, minutes: int = 0) -> typing.Self: ...

    @typing.override
    @classmethod
    async def from_native(
        cls,
        context: "Context",
        value: typing.Optional[datetime.timezone] = None,
        /,
        *,
        hours: int = 0,
        minutes: int = 0,
    ) -> typing.Self:
        if value is None:
            return await cls(hours=hours, minutes=minutes).validate()

        offset = value.utcoffset(None)
        hour, minute = divmod(offset.total_seconds() // 60, 60)

        return await cls(hours=int(hour), minutes=int(minute)).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> datetime.timezone:
        await self.validate()

        return datetime.timezone(offset=datetime.timedelta(hours=self.hours, minutes=self.minutes))

    @typing.override
    async def validate(self) -> typing.Self:
        if not isinstance(self.hours, int):
            msg = f"Expected hours of type 'int', received '{type(self.hours).__name__}'."
            raise TypeError(msg)

        if not isinstance(self.minutes, int):
            msg = f"Expected minutes of type 'int', received '{type(self.minutes).__name__}'."
            raise TypeError(msg)

        if not (-12 <= self.hours <= 14):
            msg = f"Hours must be between -12 and 14, received '{self.hours}'."
            raise ValueError(msg)

        if not (0 <= self.minutes <= 59):
            msg = f"Minutes must be between 0 and 59, received '{self.minutes}'."
            raise ValueError(msg)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                message.hours = reader.read_int32()
            elif field_number == 2:
                reader.expect_type(tag, WireType.VARINT)
                message.minutes = reader.read_uint32()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.hours:
            writer.write_uint32(8).write_int32(self.hours)
        if self.minutes:
            writer.write_uint32(16).write_uint32(self.minutes)

        if number:
            writer.ldelim()

        return writer.finish()



import dataclasses
import functools
import weakref

import typing_extensions as typing

from ..validators import check_display_name, check_identifier

if typing.TYPE_CHECKING:
    from ..errors import DefinedExecutionError
    from ..identifiers import FeatureIdentifier
    from .feature import Feature


@dataclasses.dataclass
class Handler:
    """Abstract base class for RPC handlers."""

    identifier: str = ""
    """Uniquely identifies the handler within the scope of the same feature."""

    display_name: str = ""
    """Human readable name of the handler."""

    description: str = ""
    """Describes the use and purpose of the handler."""

    errors: dict[str, type["DefinedExecutionError"]] = dataclasses.field(default_factory=dict)
    """
    A list of defined execution errors that can happen when accessing
    this handler.
    """

    feature: typing.Optional["Feature"] = None
    """The SiLA feature this handler was registered with."""

    def __post_init__(self) -> None:
        check_identifier(self.identifier)
        check_display_name(self.display_name)

        if self.feature is not None:
            self.add_to_feature(self.feature)

    @functools.cached_property
    def fully_qualified_identifier(self) -> "FeatureIdentifier":
        """Uniquely identifies the handler."""

        if self.feature is None:
            msg = (
                f"Unable to access fully qualified identifier of unregistered {self.__class__.__name__} "
                f"'{self.identifier}'. Add it to a feature first."
            )
            raise RuntimeError(msg)

        return self.feature.fully_qualified_identifier

    def add_to_feature(self, feature: "Feature") -> typing.Self:
        """
        Register this handler with a feature.

        Args:
          feature: The feature to add this handler to.

        Returns:
          The instance, allowing for method chaining.
        """

        self.feature = weakref.proxy(feature)

        return self



import dataclasses
import uuid

import typing_extensions as typing

from ..binary_transfer import (
    DeleteBinaryRequest,
    DownloadChunkRequest,
    GetBinaryInfoRequest,
    UploadChunkRequest,
)
from ..protobuf import Message, Reader, WireType, Writer
from .cancel_request import CancelRequest
from .command_execution_request import CommandExecutionRequest
from .command_response_request import CommandResponseRequest
from .create_binary_upload_request import CreateBinaryUploadRequest
from .metadata_request import MetadataRequest
from .property_request import PropertyRequest


@dataclasses.dataclass
class ClientMessage(Message):
    """
    Message sent from client to server through the cloud stream.

    Attributes:
      request_uuid: Unique ID used to map responses to requests.
    """

    request_uuid: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    unobservable_command_execution: typing.Optional[CommandExecutionRequest] = None
    observable_command_initiation: typing.Optional[CommandExecutionRequest] = None
    observable_command_execution_info: typing.Optional[CommandResponseRequest] = None
    observable_command_intermediate_response: typing.Optional[CommandResponseRequest] = None
    observable_command_get_response: typing.Optional[CommandResponseRequest] = None
    metadata_request: typing.Optional[MetadataRequest] = None
    unobservable_property_read: typing.Optional[PropertyRequest] = None
    observable_property_subscription: typing.Optional[PropertyRequest] = None
    cancel_observable_command_execution_info: typing.Optional[CancelRequest] = None
    cancel_observable_command_intermediate_response: typing.Optional[CancelRequest] = None
    cancel_observable_property: typing.Optional[CancelRequest] = None
    create_binary_upload_request: typing.Optional[CreateBinaryUploadRequest] = None
    delete_uploaded_binary_request: typing.Optional[DeleteBinaryRequest] = None
    upload_chunk_request: typing.Optional[UploadChunkRequest] = None
    get_binary_info_request: typing.Optional[GetBinaryInfoRequest] = None
    download_chunk_request: typing.Optional[DownloadChunkRequest] = None
    delete_downloaded_binary_request: typing.Optional[DeleteBinaryRequest] = None

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls(request_uuid="")
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message.request_uuid = reader.read_string()
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message.unobservable_command_execution = CommandExecutionRequest.decode(reader, reader.read_uint32())
            elif field_number == 3:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_initiation = CommandExecutionRequest.decode(reader, reader.read_uint32())
            elif field_number == 4:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_execution_info = CommandResponseRequest.decode(reader, reader.read_uint32())
            elif field_number == 5:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_intermediate_response = CommandResponseRequest.decode(
                    reader, reader.read_uint32()
                )
            elif field_number == 6:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_get_response = CommandResponseRequest.decode(reader, reader.read_uint32())
            elif field_number == 7:
                reader.expect_type(tag, WireType.LEN)
                message.metadata_request = MetadataRequest.decode(reader, reader.read_uint32())
            elif field_number == 8:
                reader.expect_type(tag, WireType.LEN)
                message.unobservable_property_read = PropertyRequest.decode(reader, reader.read_uint32())
            elif field_number == 9:
                reader.expect_type(tag, WireType.LEN)
                message.observable_property_subscription = PropertyRequest.decode(reader, reader.read_uint32())
            elif field_number == 10:
                reader.expect_type(tag, WireType.LEN)
                message.cancel_observable_command_execution_info = CancelRequest.decode(reader, reader.read_uint32())
            elif field_number == 11:
                reader.expect_type(tag, WireType.LEN)
                message.cancel_observable_command_intermediate_response = CancelRequest.decode(
                    reader, reader.read_uint32()
                )
            elif field_number == 12:
                reader.expect_type(tag, WireType.LEN)
                message.cancel_observable_property = CancelRequest.decode(reader, reader.read_uint32())
            elif field_number == 13:
                reader.expect_type(tag, WireType.LEN)
                message.create_binary_upload_request = CreateBinaryUploadRequest.decode(reader, reader.read_uint32())
            elif field_number == 14:
                reader.expect_type(tag, WireType.LEN)
                message.delete_uploaded_binary_request = DeleteBinaryRequest.decode(reader, reader.read_uint32())
            elif field_number == 15:
                reader.expect_type(tag, WireType.LEN)
                message.upload_chunk_request = UploadChunkRequest.decode(reader, reader.read_uint32())
            elif field_number == 16:
                reader.expect_type(tag, WireType.LEN)
                message.get_binary_info_request = GetBinaryInfoRequest.decode(reader, reader.read_uint32())
            elif field_number == 17:
                reader.expect_type(tag, WireType.LEN)
                message.download_chunk_request = DownloadChunkRequest.decode(reader, reader.read_uint32())
            elif field_number == 18:
                reader.expect_type(tag, WireType.LEN)
                message.delete_downloaded_binary_request = DeleteBinaryRequest.decode(reader, reader.read_uint32())
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.request_uuid:
            writer.write_uint32(10).write_string(self.request_uuid)

        if self.unobservable_command_execution is not None:
            self.unobservable_command_execution.encode(writer, 2)

        if self.observable_command_initiation is not None:
            self.observable_command_initiation.encode(writer, 3)

        if self.observable_command_execution_info is not None:
            self.observable_command_execution_info.encode(writer, 4)

        if self.observable_command_intermediate_response is not None:
            self.observable_command_intermediate_response.encode(writer, 5)

        if self.observable_command_get_response is not None:
            self.observable_command_get_response.encode(writer, 6)

        if self.metadata_request is not None:
            self.metadata_request.encode(writer, 7)

        if self.unobservable_property_read is not None:
            self.unobservable_property_read.encode(writer, 8)

        if self.observable_property_subscription is not None:
            self.observable_property_subscription.encode(writer, 9)

        if self.cancel_observable_command_execution_info is not None:
            self.cancel_observable_command_execution_info.encode(writer, 10)

        if self.cancel_observable_command_intermediate_response is not None:
            self.cancel_observable_command_intermediate_response.encode(writer, 11)

        if self.cancel_observable_property is not None:
            self.cancel_observable_property.encode(writer, 12)

        if self.create_binary_upload_request is not None:
            self.create_binary_upload_request.encode(writer, 13)

        if self.delete_uploaded_binary_request is not None:
            self.delete_uploaded_binary_request.encode(writer, 14)

        if self.upload_chunk_request is not None:
            self.upload_chunk_request.encode(writer, 15)

        if self.get_binary_info_request is not None:
            self.get_binary_info_request.encode(writer, 16)

        if self.download_chunk_request is not None:
            self.download_chunk_request.encode(writer, 17)

        if self.delete_downloaded_binary_request is not None:
            self.delete_downloaded_binary_request.encode(writer, 18)

        if number:
            writer.ldelim()

        return writer.finish()



import dataclasses

from ..framework import (
    Constrained,
    DefinedExecutionError,
    Element,
    Feature,
    FeatureIdentifier,
    FullyQualifiedIdentifier,
    Length,
    List,
    MaximalLength,
    Metadata,
    Pattern,
    Schema,
    Serializer,
    String,
)
from .server import Server
from .unobservable_command import UnobservableCommand
from .unobservable_property import UnobservableProperty

UnimplementedFeature = DefinedExecutionError.create(
    identifier="UnimplementedFeature",
    display_name="Unimplemented Feature",
    description="The Feature specified by the given Feature identifier is not implemented by the server.",
)


@dataclasses.dataclass
class SiLAService(Feature):
    sila2_version: str = "1.0"
    version: str = "1.0"
    maturity_level: str = "Normative"
    originator: str = "org.silastandard"
    category: str = "core"
    identifier: str = "SiLAService"
    display_name: str = "SiLA Service"
    description: str = (
        "This Feature MUST be implemented by each SiLA Server.\n\n"
        "It specifies Commands and Properties to discover the Features a SiLA Server implements as well as details"
        "about the SiLA Server, like name, type, description, vendor and UUID.\n\n"
        "Any interaction described in this feature MUST not affect the behaviour of any other Feature."
    )

    def __post_init__(self):
        UnobservableCommand(
            identifier="GetFeatureDefinition",
            display_name="Get Feature Definition",
            description=(
                "Get the Feature Definition of an implemented Feature by its fully qualified Feature Identifier. "
                "This command has no preconditions and no further dependencies and can be called at any time."
            ),
            parameters={
                "feature_identifier": Element(
                    identifier="FeatureIdentifier",
                    display_name="Feature Identifier",
                    description="The fully qualified Feature identifier for which the Feature definition shall be retrieved.",
                    data_type=Constrained.create(
                        data_type=String,
                        constraints=[FullyQualifiedIdentifier(FullyQualifiedIdentifier.Type.FEATURE_IDENTIFIER)],
                    ),
                ),
            },
            responses={
                "FeatureDefinition": Element(
                    identifier="FeatureDefinition",
                    display_name="Feature Definition",
                    description="The Feature definition in XML format (according to the Feature Definition Schema).",
                    data_type=Constrained.create(
                        data_type=String,
                        constraints=[
                            Schema(
                                type=Schema.Type.XML,
                                url="https://gitlab.com/SiLA2/sila_base/raw/master/schema/FeatureDefinition.xsd",
                            )
                        ],
                    ),
                ),
            },
            errors={UnimplementedFeature.identifier: UnimplementedFeature},
            function=self.get_feature_definition,
            feature=self,
        )

        UnobservableCommand(
            identifier="SetServerName",
            display_name="Set Server Name",
            description=(
                "Sets a human readable name to the Server Name Property. Command has no preconditions and "
                "no further dependencies and can be called at any time."
            ),
            parameters={
                "server_name": Element(
                    identifier="ServerName",
                    display_name="Server Name",
                    description="The human readable name to assign to the SiLA Server.",
                    data_type=Constrained.create(data_type=String, constraints=[MaximalLength(255)]),
                ),
            },
            function=self.set_server_name,
            feature=self,
        )

        UnobservableProperty(
            identifier="ServerName",
            display_name="Server Name",
            description=(
                "Human readable name of the SiLA Server. The name can be set using the 'Set Server Name' command."
            ),
            data_type=Constrained.create(String, [MaximalLength(255)]),
            function=self.get_server_name,
            feature=self,
        )

        UnobservableProperty(
            identifier="ServerType",
            display_name="Server Type",
            description=(
                "The type of this server. It, could be, e.g., in the case of a SiLA Device the model name. "
                "It is specified by the implementer of the SiLA Server and MAY not be unique."
            ),
            data_type=Constrained.create(String, [Pattern(r"[A-Z][a-zA-Z0-9]*")]),
            function=self.get_server_type,
            feature=self,
        )

        UnobservableProperty(
            identifier="ServerUUID",
            display_name="Server UUID",
            description=(
                "Globally unique identifier that identifies a SiLA Server. The Server UUID MUST be generated once "
                "and remain the same for all times."
            ),
            data_type=Constrained.create(
                String,
                [Length(36), Pattern(r"[0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{12}")],
            ),
            function=self.get_server_uuid,
            feature=self,
        )

        UnobservableProperty(
            identifier="ServerDescription",
            display_name="Server Description",
            description="Description of the SiLA Server. This should include the use and purpose of this SiLA Server.",
            data_type=String,
            function=self.get_server_description,
            feature=self,
        )

        UnobservableProperty(
            identifier="ServerVersion",
            display_name="Server Version",
            description=(
                'Returns the version of the SiLA Server. A "Major" and a "Minor" version number (e.g. 1.0) MUST '
                "be provided, a Patch version number MAY be provided. Optionally, an arbitrary text, separated by "
                "an underscore MAY be appended, e.g. “3.19.373_mighty_lab_devices”."
            ),
            data_type=Constrained.create(
                String, [Pattern(r"(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))?(_[_a-zA-Z0-9]+)?")]
            ),
            function=self.get_server_version,
            feature=self,
        )

        UnobservableProperty(
            identifier="ServerVendorURL",
            display_name="Server Vendor URL",
            description=(
                "Returns the URL to the website of the vendor or the website of the product of this SiLA Server. "
                "This URL SHOULD be accessible at all times.\n"
                "The URL is a Uniform Resource Locator as defined in RFC 1738."
            ),
            data_type=Constrained.create(String, [Pattern(r"https?://.+")]),
            function=self.get_vendor_url,
            feature=self,
        )

        UnobservableProperty(
            identifier="ImplementedFeatures",
            display_name="Implemented Features",
            description=(
                "Returns a list of fully qualified Feature identifiers of all implemented Features of this "
                "SiLA Server. This list SHOULD remain the same throughout the lifetime of the SiLA Server."
            ),
            data_type=List.create(
                data_type=Constrained.create(
                    data_type=String,
                    constraints=[FullyQualifiedIdentifier(FullyQualifiedIdentifier.Type.FEATURE_IDENTIFIER)],
                ),
            ),
            function=self.get_implemented_features,
            feature=self,
        )

    async def get_server_name(self, *, metadata: Metadata) -> dict[str, str]:
        assert isinstance(self.context, Server)

        return {"ServerName": self.context.name}

    def set_server_name(self, server_name: str, *, metadata: Metadata) -> None:
        assert isinstance(self.context, Server)

        self.context.name = server_name

    async def get_server_type(self, *, metadata: Metadata) -> dict[str, str]:
        assert isinstance(self.context, Server)

        return {"ServerType": self.context.type}

    async def get_server_uuid(self, *, metadata: Metadata) -> dict[str, str]:
        assert isinstance(self.context, Server)

        return {"ServerUUID": self.context.uuid}

    async def get_server_description(self, *, metadata: Metadata) -> dict[str, str]:
        assert isinstance(self.context, Server)

        return {"ServerDescription": self.context.description}

    async def get_server_version(self, *, metadata: Metadata) -> dict[str, str]:
        assert isinstance(self.context, Server)

        return {"ServerVersion": self.context.version}

    async def get_vendor_url(self, *, metadata: Metadata) -> dict[str, str]:
        assert isinstance(self.context, Server)

        return {"ServerVendorURL": self.context.vendor_url}

    async def get_implemented_features(self, *, metadata: Metadata) -> dict[str, list[str]]:
        return {"ImplementedFeatures": list(self.context.features)}

    async def get_feature_definition(self, feature_identifier: str, *, metadata: Metadata) -> dict[str, str]:
        identifier = FeatureIdentifier(feature_identifier)

        if identifier not in self.context.features:
            raise UnimplementedFeature

        feature = self.context.features[identifier]

        serializer = Serializer()
        feature.serialize(serializer)

        return {"FeatureDefinition": serializer.buffer.getvalue()}



import asyncio
import contextlib
import dataclasses
import functools
import logging
import types
import weakref

import grpc
import grpc.aio
import typing_extensions as typing

from ..framework import (
    BinaryTransferError,
    ChannelOptions,
    ClientMessage,
    CloudMetadata,
    CommandConfirmationResponse,
    CommandExecutionResponse,
    CommandExecutionUUID,
    DecodeError,
    EncodeError,
    InvalidCommandExecutionUUID,
    MetadataResponse,
    ObservableCommandResponse,
    PropertyResponse,
    Server,
    ServerMessage,
    SiLAError,
    Stream,
    UndefinedExecutionError,
    UnobservableCommandResponse,
    create_channel,
)
from .server_binary_transfer_handler import ServerBinaryTransferHandler

if typing.TYPE_CHECKING:
    from ..framework import Feature, FeatureIdentifier, Protobuf
    from .command_execution import CommandExecution


@dataclasses.dataclass
class CloudServerConfig:
    """Configuration to run a SiLA 2 cloud server."""

    hostname: str = "localhost"
    """The target hostname to connect to."""

    port: int = 50_001
    """The target port to connect to."""

    tls: bool = True
    """
    Whether to run a secure/TLS channel or a plaintext channel (i.e.
    no TLS), defaults to run with TLS encryption.
    """

    root_certificates: typing.Optional[bytes] = None
    """
    The PEM-encoded root certificates as a byte string, or None to
    retrieve them from a default location chosen by gRPC runtime.
    """

    certificate_chain: typing.Optional[bytes] = None
    """
    The PEM-encoded certificate chain as a byte string to use or None
    if no certificate chain should be used.
    """

    private_key: typing.Optional[bytes] = None
    """
    The PEM-encoded private key as a byte string, or None if no
    private key should be used.
    """

    reconnect_delay: float = 10000.0
    """
    The time in ms to wait before reconnecting the channel after an
    error occurs.
    """

    options: ChannelOptions = dataclasses.field(default_factory=ChannelOptions)
    """
    Additional options for the underlying gRPC connection.
    """


class CloudServer(Server):
    """
    A SiLA 2 cloud server capable of using the server-initiated connection method.

    Args:
      target: The target address to connect to.
      options: Additional options for the connection.
    """

    def __init__(self, config: typing.Optional[typing.Union[dict, CloudServerConfig]] = None, /):
        self._context: Server = weakref.proxy(self)
        super().__init__()
        if isinstance(config, dict):
            import warnings

            warnings.warn(
                "Providing CloudServerConfig as a dictionary is deprecated and will be removed in a future release. "
                "Please provide a CloudServerConfig instance instead.",
                category=DeprecationWarning,
            )
            self._config = CloudServerConfig(**config)
        else:
            self._config = config or CloudServerConfig()
        self._channel: typing.Optional[grpc.aio.Channel] = None
        self._responses = Stream[ServerMessage]()
        self._shutdown = asyncio.Event()
        self._rpc_handler: typing.Optional[asyncio.Task] = None
        self._tasks: dict[str, asyncio.Task] = {}

    @property
    def context(self) -> typing.Optional[Server]:
        """The context."""

        return self._context

    @context.setter
    def context(self, context: Server) -> None:
        self._context = weakref.proxy(context)

    @property
    @typing.override
    def protobuf(self) -> "Protobuf":
        return self._context._protobuf

    @property
    @typing.override
    def features(self) -> dict["FeatureIdentifier", "Feature"]:
        return self._context._features

    @property
    @typing.override
    def command_executions(self) -> dict[str, "CommandExecution"]:
        return self._context._command_executions

    @property
    @typing.override
    def binary_transfer_handler(self) -> "ServerBinaryTransferHandler":
        return self._context._binary_transfer_handler

    @property
    def tasks(self) -> dict[str, asyncio.Task]:
        """The currently ongoing tasks."""

        return self._tasks

    @property
    def logger(self) -> logging.Logger:
        """A python logger instance."""

        return logging.getLogger(__name__)

    @typing.override
    async def _start(self) -> None:
        await super()._start()

        try:
            self._responses = Stream[ServerMessage]()
            target = f"{self._config.hostname}:{self._config.port}"
            self._channel = await functools.partial(create_channel, **dataclasses.asdict(self._config))(target)
            self.logger.info(f"{self.__class__.__name__} connected to '{target}'.")

            listen = self._channel.stream_stream(
                method="/sila2.org.silastandard.CloudClientEndpoint/ConnectSiLAServer",
                request_serializer=self._serialize_request,
                response_deserializer=self._deserialize_response,
            )
            request_iterator: grpc.aio.StreamStreamCall[ServerMessage, ClientMessage] = listen(
                request_iterator=self._responses, wait_for_ready=True
            )
            self._rpc_handler = asyncio.create_task(self.rpc_handler(request_iterator))
        except grpc.aio.AioRpcError as error:
            self.logger.error(error)
            await self.stop()

    async def rpc_handler(self, request_iterator: grpc.aio.StreamStreamCall[ServerMessage, ClientMessage]) -> None:
        """
        Handle the bidirectional gRPC call between cloud server and client.

        Args:
          request_iterator: The cloud client's client message.
        """

        try:
            async for response in request_iterator:
                await self.receive(response)
        except asyncio.CancelledError:
            pass
        except grpc.aio.AioRpcError as error:
            self.logger.error(error)

            await self.stop()
            await asyncio.sleep(self._config.reconnect_delay / 1000)
            await self.start()

    @typing.override
    async def _stop(self, grace: typing.Optional[float] = None) -> None:
        await super()._stop(grace)

        for task in self._tasks.values():
            task.cancel()

        if self._rpc_handler and self._rpc_handler != asyncio.current_task():
            self._rpc_handler.cancel()
            self._rpc_handler = None

        if self._channel:
            await self._channel.close(grace=grace)
            self._channel = None

    async def receive(self, message: ClientMessage) -> None:
        """
        Receive and handle a client message.

        Args:
          message: The cloud client's request to handle.
        """

        self.logger.debug("> %s", message)

        message_type: str = ""
        coroutine: typing.Optional[types.CoroutineType[typing.Any, typing.Any, None]] = None

        if message.unobservable_command_execution is not None:
            message_type = "unobservable_command_execution"
            coroutine = self.unobservable_command_execution(
                message.request_uuid,
                message.unobservable_command_execution.fully_qualified_command_id,
                message.unobservable_command_execution.command_parameter.parameters,
                self._parse_metadata(message.unobservable_command_execution.command_parameter.metadata),
            )

        elif message.observable_command_initiation is not None:
            message_type = "observable_command_initiation"
            coroutine = self.observable_command_initiation(
                message.request_uuid,
                message.observable_command_initiation.fully_qualified_command_id,
                message.observable_command_initiation.command_parameter.parameters,
                self._parse_metadata(message.observable_command_initiation.command_parameter.metadata),
            )

        elif message.observable_command_execution_info is not None:
            message_type = "observable_command_execution_info"
            coroutine = self.observable_command_execution_info_subscription(
                message.request_uuid,
                message.observable_command_execution_info.command_execution_uuid,
            )

        elif message.observable_command_intermediate_response is not None:
            message_type = "observable_command_intermediate_response"
            coroutine = self.observable_command_intermediate_response_subscription(
                message.request_uuid,
                message.observable_command_intermediate_response.command_execution_uuid,
            )

        elif message.observable_command_get_response is not None:
            message_type = "observable_command_get_response"
            coroutine = self.observable_command_get_response(
                message.request_uuid,
                message.observable_command_get_response.command_execution_uuid,
            )

        elif message.metadata_request is not None:
            affected_calls = []

            with contextlib.suppress(ValueError):
                metadata = self._context.get_metadata(message.metadata_request.fully_qualified_metadata_id)
                affected_calls = metadata.affects

            msg = ServerMessage(
                request_uuid=message.request_uuid,
                get_fcp_affected_by_metadata_response=MetadataResponse(affected_calls),
            )
            await self.respond(msg)

        elif message.unobservable_property_read is not None:
            message_type = "unobservable_property_read"
            coroutine = self.unobservable_property_read(
                message.request_uuid,
                message.unobservable_property_read.fully_qualified_property_id,
                self._parse_metadata(message.unobservable_property_read.metadata),
            )

        elif message.observable_property_subscription is not None:
            message_type = "observable_property_subscription"
            coroutine = self.observable_property_subscription(
                message.request_uuid,
                message.observable_property_subscription.fully_qualified_property_id,
                self._parse_metadata(message.observable_property_subscription.metadata),
            )

        elif message.cancel_observable_command_execution_info is not None:
            await self.cancel_observable_command_execution_info_subscription(
                message.request_uuid,
            )

        elif message.cancel_observable_command_intermediate_response is not None:
            await self.cancel_observable_command_intermediate_response_subscription(
                message.request_uuid,
            )

        elif message.cancel_observable_property is not None:
            await self.cancel_observable_property_subscription(message.request_uuid)

        elif message.create_binary_upload_request is not None:
            try:
                response = await self._binary_transfer_handler.create_binary(
                    message.create_binary_upload_request.create_binary_request,
                    self._parse_metadata(message.create_binary_upload_request.metadata),
                )
                msg = ServerMessage(request_uuid=message.request_uuid, create_binary_response=response)
                await self.respond(msg)
            except BinaryTransferError as error:
                msg = ServerMessage(request_uuid=message.request_uuid, binary_transfer_error=error)
                await self.respond(msg)
            except SiLAError as error:
                msg = ServerMessage(request_uuid=message.request_uuid, command_error=error)
                await self.respond(msg)

        elif message.upload_chunk_request is not None:
            try:
                response = await self._binary_transfer_handler.upload_chunk(message.upload_chunk_request)
                msg = ServerMessage(request_uuid=message.request_uuid, upload_chunk_response=response)
                await self.respond(msg)
            except BinaryTransferError as error:
                msg = ServerMessage(request_uuid=message.request_uuid, binary_transfer_error=error)
                await self.respond(msg)

        elif message.delete_uploaded_binary_request is not None:
            try:
                response = await self._binary_transfer_handler.delete_binary(message.delete_uploaded_binary_request)
                msg = ServerMessage(request_uuid=message.request_uuid, delete_binary_response=response)
                await self.respond(msg)
            except BinaryTransferError as error:
                msg = ServerMessage(request_uuid=message.request_uuid, binary_transfer_error=error)
                await self.respond(msg)

        elif message.get_binary_info_request is not None:
            try:
                response = await self._binary_transfer_handler.get_binary_info(message.get_binary_info_request)
                msg = ServerMessage(request_uuid=message.request_uuid, get_binary_info_response=response)
                await self.respond(msg)
            except BinaryTransferError as error:
                msg = ServerMessage(request_uuid=message.request_uuid, binary_transfer_error=error)
                await self.respond(msg)

        elif message.download_chunk_request is not None:
            try:
                response = await self._binary_transfer_handler.download_chunk(message.download_chunk_request)
                msg = ServerMessage(request_uuid=message.request_uuid, download_chunk_response=response)
                await self.respond(msg)
            except BinaryTransferError as error:
                msg = ServerMessage(request_uuid=message.request_uuid, binary_transfer_error=error)
                await self.respond(msg)

        elif message.delete_downloaded_binary_request is not None:
            try:
                response = await self._binary_transfer_handler.delete_binary(message.delete_downloaded_binary_request)
                msg = ServerMessage(request_uuid=message.request_uuid, delete_binary_response=response)
                await self.respond(msg)
            except BinaryTransferError as error:
                msg = ServerMessage(request_uuid=message.request_uuid, binary_transfer_error=error)
                await self.respond(msg)

        if coroutine is not None:
            task_name = f"{message.request_uuid}_{message_type}"
            self._tasks[task_name] = asyncio.create_task(coroutine, name=task_name)
            self._tasks[task_name].add_done_callback(lambda _: self._tasks.pop(task_name))

    async def respond(self, message: ServerMessage) -> None:
        """
        Respond with a server message.

        Args:
          message: The cloud client's response to answer with.
        """

        self.logger.debug("< %s", message)
        self._responses.next(message)

    async def unobservable_command_execution(
        self, request_uuid: str, identifier: str, parameters: bytes, metadata: dict[str, bytes]
    ) -> None:
        """
        Execute an unobservable command.

        Args:
          request_uuid: A unique identifier for the request.
          identifier: The fully qualified identifier of the unobservable
            command to be executed.
          parameters: Input parameters provided to command execution.
          metadata: Additional metadata required for processing the
            request.
        """

        try:
            unobservable_command = self._context.get_unobservable_command(identifier)
        except ValueError as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=UndefinedExecutionError(str(error)))
            await self.respond(msg)
            return

        try:
            responses = await unobservable_command.execute(parameters, metadata)
        except SiLAError as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=error)
            await self.respond(msg)
        else:
            msg = ServerMessage(
                request_uuid=request_uuid,
                unobservable_command_response=UnobservableCommandResponse(response=responses),
            )
            await self.respond(msg)

    async def observable_command_initiation(
        self, request_uuid: str, identifier: str, parameters: bytes, metadata: dict[str, bytes]
    ) -> None:
        """
        Initiate an observable command.

        Args:
          request_uuid: A unique identifier for the request.
          identifier: The fully qualified identifier of the observable
            command to be initiated.
          parameters: Input parameters provided to command execution.
          metadata: Additional metadata required for processing the
            request.
        """

        try:
            observable_command = self._context.get_observable_command(identifier)
        except ValueError as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=UndefinedExecutionError(str(error)))
            await self.respond(msg)
            return

        try:
            command_confirmation = await observable_command.initiate(parameters, metadata)
        except SiLAError as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=error)
            await self.respond(msg)
        else:
            msg = ServerMessage(
                request_uuid=request_uuid,
                observable_command_confirmation=CommandConfirmationResponse(command_confirmation=command_confirmation),
            )
            await self.respond(msg)

    async def observable_command_execution_info_subscription(
        self, request_uuid: str, command_execution_uuid: CommandExecutionUUID
    ) -> None:
        """
        Subscribe the execution info of an observable command execution.

        Args:
          request_uuid: A unique identifier for the request.
          command_execution_uuid: The unique identifier of the observable
            command execution to subscribe info for.
        """

        try:
            command_execution = self._context.get_command_execution(command_execution_uuid.value)
        except InvalidCommandExecutionUUID as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=error)
            await self.respond(msg)
            return

        observable_command = command_execution.command
        async for status in observable_command.subscribe_status(command_execution.command_execution_uuid):
            msg = ServerMessage(
                request_uuid=request_uuid,
                observable_command_execution_info=CommandExecutionResponse(
                    command_execution_uuid=CommandExecutionUUID(value=command_execution.command_execution_uuid),
                    execution_info=status,
                ),
            )
            await self.respond(msg)

    async def cancel_observable_command_execution_info_subscription(self, request_uuid: str) -> None:
        """
        Cancel the subscription of command execution info.

        Args:
          request_uuid: The unique identifier associated with the
            subscription request.
        """

        if f"{request_uuid}_observable_command_execution_info" in self._tasks:
            self._tasks.pop(f"{request_uuid}_observable_command_execution_info").cancel()

    async def observable_command_intermediate_response_subscription(
        self, request_uuid: str, command_execution_uuid: CommandExecutionUUID
    ) -> None:
        """
        Subscribe the intermediate responses of an observable command execution.

        Args:
          request_uuid: A unique identifier for the request.
          command_execution_uuid: The unique identifier of the observable
            command execution to subscribe responses for.
        """

        try:
            command_execution = self._context.get_command_execution(command_execution_uuid.value)
        except InvalidCommandExecutionUUID as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=error)
            await self.respond(msg)
            return

        observable_command = command_execution.command
        async for responses in observable_command.subscribe_intermediate(command_execution.command_execution_uuid):
            msg = ServerMessage(
                request_uuid=request_uuid,
                observable_command_intermediate_response=ObservableCommandResponse(
                    command_execution_uuid=CommandExecutionUUID(value=command_execution.command_execution_uuid),
                    response=responses,
                ),
            )
            await self.respond(msg)

    async def cancel_observable_command_intermediate_response_subscription(self, request_uuid: str) -> None:
        """
        Cancel the subscription of command intermediate responses.

        Args:
          request_uuid: The unique identifier associated with the
            subscription request.
        """

        if f"{request_uuid}_observable_command_intermediate_response" in self._tasks:
            self._tasks.pop(f"{request_uuid}_observable_command_intermediate_response").cancel()

    async def observable_command_get_response(
        self, request_uuid: str, command_execution_uuid: CommandExecutionUUID
    ) -> None:
        """
        Get the response of an observable command.

        Args:
          request_uuid: A unique identifier for the request.
          command_execution_uuid: The unique identifier of the observable
            command execution to get responses for.
        """

        try:
            command_execution = self._context.get_command_execution(command_execution_uuid.value)
        except InvalidCommandExecutionUUID as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=error)
            await self.respond(msg)
            return

        try:
            observable_command = command_execution.command
            response = await observable_command.get_result(command_execution.command_execution_uuid)

            msg = ServerMessage(
                request_uuid=request_uuid,
                observable_command_response=ObservableCommandResponse(
                    command_execution_uuid=CommandExecutionUUID(value=command_execution.command_execution_uuid),
                    response=response,
                ),
            )
            await self.respond(msg)
        except SiLAError as error:
            msg = ServerMessage(request_uuid=request_uuid, command_error=error)
            await self.respond(msg)

    async def unobservable_property_read(self, request_uuid: str, identifier: str, metadata: dict[str, bytes]) -> None:
        """
        Read an unobservable property.

        Args:
          request_uuid: A unique identifier for the request.
          identifier: The fully qualified identifier of the unobservable
            property to be read.
          metadata: Additional metadata required for processing the
            request.
        """

        try:
            unobservable_property = self._context.get_unobservable_property(identifier)
        except ValueError as error:
            msg = ServerMessage(request_uuid=request_uuid, property_error=UndefinedExecutionError(str(error)))
            await self.respond(msg)
            return

        try:
            responses = await unobservable_property.read(metadata)
        except SiLAError as error:
            msg = ServerMessage(request_uuid=request_uuid, property_error=error)
            await self.respond(msg)
        else:
            msg = ServerMessage(
                request_uuid=request_uuid,
                unobservable_property_value=PropertyResponse(value=responses),
            )
            await self.respond(msg)

    async def observable_property_subscription(
        self, request_uuid: str, identifier: str, metadata: dict[str, bytes]
    ) -> None:
        """
        Subscribe an observable property.

        Args:
          request_uuid: A unique identifier for the request.
          identifier: The fully qualified identifier of the observable
            property to be subscribed.
          metadata: Additional metadata required for processing the
            request.
        """

        try:
            observable_property = self._context.get_observable_property(identifier)
        except ValueError as error:
            msg = ServerMessage(request_uuid=request_uuid, property_error=UndefinedExecutionError(str(error)))
            await self.respond(msg)
            return

        try:
            async for responses in observable_property.subscribe(metadata):
                message = ServerMessage(
                    request_uuid=request_uuid,
                    observable_property_value=PropertyResponse(value=responses),
                )
                await self.respond(message)

        except SiLAError as error:
            message = ServerMessage(request_uuid=request_uuid, property_error=error)
            await self.respond(message)

    async def cancel_observable_property_subscription(self, request_uuid: str) -> None:
        """
        Cancel the subscription of an observable property.

        Args:
          request_uuid: The unique identifier associated with the
            subscription request.
        """

        if f"{request_uuid}_observable_property_subscription" in self._tasks:
            self._tasks.pop(f"{request_uuid}_observable_property_subscription").cancel()

    def _serialize_request(self, message: ServerMessage) -> bytes:
        try:
            return message.encode()
        except EncodeError as error:
            self.logger.error(error)
            return b""

    def _deserialize_response(self, buffer: bytes) -> ClientMessage:
        try:
            return ClientMessage.decode(buffer)
        except DecodeError as error:
            self.logger.error(error)
            return ClientMessage()

    def _parse_metadata(self, metadata: list[CloudMetadata]) -> dict[str, bytes]:
        return {
            f"sila-{metadata.fully_qualified_metadata_id.replace('/', '-').lower()}-bin": metadata.value
            for metadata in metadata
        }



import asyncio
import functools
import uuid
import weakref

import typing_extensions as typing

from sila import datetime

from .. import framework
from ..framework import (
    CommandConfirmation,
    CommandExecutionInfo,
    CommandExecutionNotFinished,
    CommandExecutionStatus,
    CommandExecutionUUID,
    Duration,
    MetadataIdentifier,
    Native,
    Real,
    Stream,
)

if typing.TYPE_CHECKING:
    from .observable_command import ObservableCommand


T = typing.TypeVar("T", bound=Native)


class CommandExecution(framework.CommandExecution, typing.Generic[T]):
    """Tracks the progress of an observable command execution."""

    def __init__(
        self,
        command: "ObservableCommand",
        parameters: typing.Optional[dict[str, Native]] = None,
        metadata: typing.Optional[dict[MetadataIdentifier, Native]] = None,
        lifetime: typing.Optional[datetime.timedelta] = None,
    ) -> None:
        self._command = weakref.proxy(command)
        self._execute = functools.partial(command.execute, parameters or {}, metadata or {}, command_execution=self)
        self._lifetime = lifetime
        self._result: typing.Optional[bytes] = None
        self._exception: typing.Optional[BaseException] = None
        self._task: typing.Optional[asyncio.Task] = None
        self._running: bool = False

        self.command_execution_uuid = str(uuid.uuid4())
        self.execution_info = Stream[CommandExecutionInfo](maxsize=1)
        self.intermediate_responses = Stream[T](maxsize=1)

        self.execution_info.next(
            CommandExecutionInfo(
                status=CommandExecutionStatus.WAITING,
                progress=Real(0.0),
            )
        )

    @property
    @typing.override
    def command(self) -> "ObservableCommand":
        return self._command

    @property
    def command_confirmation(self) -> CommandConfirmation:
        """A command confirmation message is returned to identify the command execution."""

        return CommandConfirmation(
            command_execution_uuid=CommandExecutionUUID(value=self.command_execution_uuid),
            lifetime_of_execution=Duration.from_total_seconds(self._lifetime.total_seconds())
            if self._lifetime is not None
            else None,
        )

    def execute(self) -> None:
        """Execute the observable command in a background task."""

        if self._task is not None:
            msg = "The command is already executing."
            raise RuntimeError(msg)

        self._task = asyncio.create_task(self._execute())

    def done(self) -> bool:
        """Whether the command execution has a result or an exception set."""

        return self._result is not None or self._exception is not None

    def result(self) -> bytes:
        """
        Return the result of the command execution.

        Returns:
          The result value.

        Raises:
          CommandExecutionNotFinished: If the result isn't yet available.
          BaseException: If the command execution is done and has an
            exception set by the `set_exception()` method.
        """

        if self._exception is not None:
            raise self._exception

        if self._result is None:
            msg = "Result is not ready."
            raise CommandExecutionNotFinished(msg)

        return self._result

    async def set_result(self, result: bytes) -> None:
        """
        Mark the command execution as done and set its result.

        Args:
          result: The result to set.
        """

        if not self._running:
            self.execution_info.next(
                CommandExecutionInfo(
                    status=CommandExecutionStatus.RUNNING,
                    progress=Real(1.0),
                    remaining_time=Duration(),
                )
            )
            await asyncio.sleep(0)

        self._result = result
        self.execution_info.next(
            CommandExecutionInfo(
                status=CommandExecutionStatus.FINISHED_SUCCESSFULLY,
                progress=Real(1.0),
                remaining_time=Duration(),
            )
        )
        self.execution_info.close()
        self.intermediate_responses.close()

    async def set_exception(self, exception: BaseException) -> None:
        """
        Mark the command execution as done and set an exception.

        Args:
          exception: The exception to raise.
        """

        if not self._running:
            self.execution_info.next(
                CommandExecutionInfo(
                    status=CommandExecutionStatus.RUNNING,
                    progress=Real(1.0),
                    remaining_time=Duration(),
                )
            )
            await asyncio.sleep(0)

        self._exception = exception
        self.execution_info.next(
            CommandExecutionInfo(
                status=CommandExecutionStatus.FINISHED_WITH_ERROR,
                progress=Real(1.0),
                remaining_time=Duration(),
            )
        )
        self.execution_info.close()
        self.intermediate_responses.close()

    def update_execution_info(
        self,
        progress: typing.Optional[float] = None,
        remaining_time: typing.Optional[datetime.timedelta] = None,
        updated_lifetime: typing.Optional[datetime.timedelta] = None,
    ) -> None:
        """
        Update the current execution info of the command execution.

        Args:
          progress: The estimated progress in percent (0 ... 100%).
          remaining_time: The estimated remaining execution time.
          updated_lifetime: The duration during which this execution is
            valid.
        """

        if self._lifetime is not None and updated_lifetime is not None and updated_lifetime > self._lifetime:
            self._lifetime = updated_lifetime

        execution_info = CommandExecutionInfo(
            status=CommandExecutionStatus.RUNNING,
            progress=Real(progress) if progress is not None else None,
            remaining_time=Duration.from_total_seconds(remaining_time.total_seconds())
            if remaining_time is not None
            else None,
            updated_lifetime=Duration.from_total_seconds(self._lifetime.total_seconds())
            if self._lifetime is not None
            else None,
        )
        self.execution_info.next(execution_info)
        self._running = True

    def send_intermediate_responses(self, intermediate_responses: T) -> None:
        """Send intermediate responses to the client."""

        self.intermediate_responses.next(intermediate_responses)

    @typing.override
    def cancel(self) -> None:
        if self._task is not None:
            self._task.cancel()

        self.execution_info.close()
        self.intermediate_responses.close()



import contextlib
import ipaddress
import logging
import socket
import weakref

import typing_extensions as typing
import zeroconf.asyncio

if typing.TYPE_CHECKING:
    from .server import Server


class Discovery:
    """
    Promotes the provided server on the network.

    In order to provide a true zero-configuration experience, this
    implements multicast DNS (mDNS) and DNS-based Service Discovery
    (DNS-SD). It broadcasts the socket address on which the grpc
    server is available, some metadata about the server and
    optionally the server certificate.

    Args:
      server: The server to promote on the network.
    """

    def __init__(self, server: "Server"):
        self._server: "Server" = weakref.proxy(server)

        self.mdns = zeroconf.asyncio.AsyncZeroconf(interfaces=zeroconf.InterfaceChoice.Default)

    @property
    def logger(self) -> logging.Logger:
        """A python logger instance."""

        return logging.getLogger(__name__)

    async def start(self) -> None:
        """Start broadcasting the service on mDNS."""

        await self._server.wait_for_ready()
        service = Discovery.create_service(self._server)

        await self.mdns.zeroconf.async_wait_for_start()
        await self.mdns.async_register_service(service)
        self.logger.info("Registered service for discovery.")

    async def stop(self) -> None:
        """Stop broadcasting the service on mDNS."""

        await self.mdns.async_unregister_all_services()
        await self.mdns.async_close()
        self.logger.info("Removed service from discovery.")

    @classmethod
    def find_ip_address(cls, address: str) -> str:
        """
        Return a valid ip address for a given hostname or address.

        Args:
            address: A hostname or address, e.g. `localhost` or `0.0.0.0`.

        Returns:
            A valid ip address that represents the given address.
        """
        if address == "0.0.0.0":
            ip_address = "127.0.0.1"
            with contextlib.suppress(OSError):
                ip_address = socket.gethostbyname(socket.gethostname())

            if ip_address == "127.0.0.1":
                connection = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                connection.settimeout(0)
                with contextlib.suppress(OSError):
                    connection.connect(("8.8.8.8", 80))
                ip_address = next(iter(connection.getsockname()), ip_address)

            return ip_address

        try:
            return ipaddress.ip_address(address).exploded
        except ValueError:
            try:
                return socket.gethostbyname(address)
            except OSError:
                return "127.0.0.1"

    @classmethod
    def create_service(cls, server: "Server") -> zeroconf.asyncio.AsyncServiceInfo:
        """Create zeroconf service info from the given server."""

        properties = {
            "version": server.version.encode("utf-8")[:247],
            "server_name": server.name.encode("utf-8")[:243],
            "description": server.description.encode("utf-8")[:243],
        }

        if server._config.tls:
            certificate = server._config.certificate_chain
            if certificate:
                properties.update({f"ca{i}": line for i, line in enumerate(certificate.splitlines(keepends=False))})

        hostname, _, port = server._address.rpartition(":")

        return zeroconf.asyncio.AsyncServiceInfo(
            type_="_sila._tcp.local.",
            name=f"{server.uuid}._sila._tcp.local.",
            parsed_addresses=[Discovery.find_ip_address(hostname)],
            port=int(port),
            properties=properties,
        )



import collections.abc
import dataclasses
import enum

import typing_extensions as typing

from ..data_types.string import String
from ..fdl import Deserializer, Serializer
from ..identifiers import (
    CommandIdentifier,
    DataTypeIdentifier,
    ErrorIdentifier,
    FeatureIdentifier,
    IntermediateResponseIdentifier,
    MetadataIdentifier,
    ParameterIdentifier,
    PropertyIdentifier,
    ResponseIdentifier,
)
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List


class Identifier(str, enum.Enum):
    """
    Enum class representing various SiLA identifier types.
    """

    FEATURE_IDENTIFIER = "FeatureIdentifier"
    """Identifier for a feature."""

    COMMAND_IDENTIFIER = "CommandIdentifier"
    """Identifier for a command."""

    COMMAND_PARAMETER_IDENTIFIER = "ParameterIdentifier"
    """Identifier for a command parameter."""

    COMMAND_RESPONSE_IDENTIFIER = "ResponseIdentifier"
    """Identifier for a command response."""

    INTERMEDIATE_COMMAND_RESPONSE_IDENTIFIER = "IntermediateResponseIdentifier"
    """Identifier for an intermediate command response."""

    DEFINED_EXECUTION_ERROR_IDENTIFIER = "ErrorIdentifier"
    """Identifier for a defined execution error."""

    PROPERTY_IDENTIFIER = "PropertyIdentifier"
    """Identifier for a property."""

    DATA_TYPE_IDENTIFIER = "DataTypeIdentifier"
    """Identifier for a data type."""

    METADATA_IDENTIFIER = "MetadataIdentifier"
    """Identifier for metadata."""


@dataclasses.dataclass
class FullyQualifiedIdentifier(Constraint[str]):
    """
    A Fully Qualified Identifier Constraint specifies the content of
    the SiLA String Type to be a Fully Qualified Identifier and
    indicates the type of the identifier. Note that this is
    comparable to a Pattern Constraint; the content is not required
    to actually identify something, it just has to be a semantically
    correct Fully Qualified Identifier.

    Args:
      value: The specific identifier type (e.g., 'FeatureIdentifier',
        'CommandIdentifier', etc.) that the string must match.
    """

    Type: typing.ClassVar[type[Identifier]] = Identifier
    """
    Enum class representing various SiLA identifier types.
    """

    value: typing.Literal[
        "FeatureIdentifier",
        "CommandIdentifier",
        "ParameterIdentifier",
        "ResponseIdentifier",
        "IntermediateResponseIdentifier",
        "ErrorIdentifier",
        "PropertyIdentifier",
        "DataTypeIdentifier",
        "MetadataIdentifier",
    ]
    """
    The specific identifier type (e.g., 'FeatureIdentifier',
    'CommandIdentifier', etc.) that the string must match.
    """

    def __post_init__(self):
        self.value = self.value.value if isinstance(self.value, Identifier) else self.value

        try:
            self.__validate = {
                Identifier.FEATURE_IDENTIFIER: FeatureIdentifier,
                Identifier.COMMAND_IDENTIFIER: CommandIdentifier,
                Identifier.COMMAND_PARAMETER_IDENTIFIER: ParameterIdentifier,
                Identifier.COMMAND_RESPONSE_IDENTIFIER: ResponseIdentifier,
                Identifier.INTERMEDIATE_COMMAND_RESPONSE_IDENTIFIER: IntermediateResponseIdentifier,
                Identifier.DEFINED_EXECUTION_ERROR_IDENTIFIER: ErrorIdentifier,
                Identifier.PROPERTY_IDENTIFIER: PropertyIdentifier,
                Identifier.DATA_TYPE_IDENTIFIER: DataTypeIdentifier,
                Identifier.METADATA_IDENTIFIER: MetadataIdentifier,
            }[Identifier(self.value)]
        except ValueError:
            raise ValueError(f"Identifier type must be valid type, received '{self.value}'.") from None

    async def validate(self, value: str) -> bool:
        if not isinstance(value, str):
            raise TypeError(f"Expected value of type 'str', received '{type(value).__name__}'.")

        try:
            self.__validate(value)
        except Exception:
            raise ValueError(f"Expected value with format for a '{self.value}', received '{value}'.") from None

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.write_str("FullyQualifiedIdentifier", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, String):
            raise ValueError(f"Expected constraint's data type to be 'String', received '{data_type.__name__}'.")

        yield from deserializer.read_start_element(name="FullyQualifiedIdentifier")
        value = yield from deserializer.read_str()
        try:
            value = Identifier(value).value
        except ValueError:
            raise ValueError(f"Expected a valid 'FullyQualifiedIdentifier' value, received '{value}'.") from None
        yield from deserializer.read_end_element(name="FullyQualifiedIdentifier")

        return cls(value)



import dataclasses

import typing_extensions as typing

from ..data_types import Duration
from ..protobuf import Message, Reader, WireType, Writer
from .command_execution_uuid import CommandExecutionUUID


@dataclasses.dataclass
class CommandConfirmation(Message):
    """
    A command confirmation message is returned to identify the command execution.

    Attributes:
      command_execution_uuid: The unique identifier of a command
        execution. It is unique within one instance of a SiLA server
        and its lifetime.
      lifetime_of_execution: The duration during which a command
        execution UUID is valid. The lifetime of execution is always
        a relative duration with respect to the point in time the
        SiLA server initiated the response to the SiLA client.
    """

    command_execution_uuid: CommandExecutionUUID = dataclasses.field(default_factory=CommandExecutionUUID)
    lifetime_of_execution: typing.Optional[Duration] = None

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message.command_execution_uuid = CommandExecutionUUID.decode(reader, reader.read_uint32())
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message.lifetime_of_execution = Duration.decode(reader, reader.read_uint32())
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        self.command_execution_uuid.encode(writer, 1)
        if self.lifetime_of_execution is not None:
            self.lifetime_of_execution.encode(writer, 2)

        if number:
            writer.ldelim()

        return writer.finish()



import collections.abc
import dataclasses

import typing_extensions as typing

from .property import Property


@dataclasses.dataclass
class ObservableProperty(Property):
    """A property describes certain aspects of a SiLA server that do not require an action on the SiLA server."""

    observable: bool = dataclasses.field(init=False, repr=False, default=True)

    def subscribe(self, metadata: typing.Optional[dict] = None) -> collections.abc.AsyncIterator[bytes]:
        """
        Subscribe to value changes of the property.

        Args:
          metadata: Additional metadata sent from client to server.

        Yields:
          The current value of the property.

        Raises:
          NoMetadataAllowed: If providing metadata is not allowed.
          InvalidMetadata: If metadata is missing or invalid.
          DefinedExecutionError: If property access results in a defined
            execution error
          UnefinedExecutionError: If property access results in an
            undefined execution error.
        """

        raise NotImplementedError



import dataclasses

import typing_extensions as typing

from ..protobuf import Reader, WireType, Writer
from .data_type import BasicType

if typing.TYPE_CHECKING:
    from ..common import Context


@dataclasses.dataclass
class Boolean(BasicType[bool]):
    """
    Represents a boolean value.

    Attributes:
      value: The encapsulated boolean value. Defaults to false.
    """

    value: bool = False

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[bool] = None, /) -> typing.Self:
        if value is None:
            return await cls().validate()

        return await cls(value=value).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> bool:
        await self.validate()

        return self.value

    @typing.override
    async def validate(self) -> typing.Self:
        if not isinstance(self.value, bool):
            msg = f"Expected value of type 'bool', received '{type(self.value).__name__}'."
            raise TypeError(msg)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                message.value = reader.read_bool()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.value:
            writer.write_uint32(8).write_bool(self.value)

        if number:
            writer.ldelim()

        return writer.finish()



import dataclasses
import inspect

import grpc
import grpc.aio
import typing_extensions as typing

from .. import framework
from ..framework import (
    Constrained,
    DecodeError,
    DefinedExecutionError,
    Element,
    Handler,
    List,
    Message,
    MetadataIdentifier,
    Native,
    Reader,
    String,
    WireType,
    Writer,
)

if typing.TYPE_CHECKING:
    from ..framework import Feature


@dataclasses.dataclass
class Metadata(framework.Metadata, Message):
    """Aditional information the server expects to receive from a client."""

    function: typing.ClassVar[typing.Callable] = lambda *args: ...
    """The implementation which is executed by the RPC handler."""

    async def intercept(self, context: Handler) -> dict[MetadataIdentifier, Native]:
        """
        Intercept the current handler execution with this metadata.

        Args:
          context: The target affected by the interception.

        Raises:
          NoMetadataAllowed: If providing metadata is not allowed.
          InvalidMetadata: If metadata is missing or invalid.
          DefinedExecutionError: If execution the metadata's interceptor
            results in a defined execution error
          UnefinedExecutionError: If execution the metadata's interceptor
            results in an undefined execution error.
        """

        assert self.feature is not None

        native = await self.to_native(context)

        try:
            response = self.function(native)

            if inspect.isawaitable(response):
                response = await response

        except DefinedExecutionError as error:
            error = error.with_feature(self.feature.fully_qualified_identifier)
            raise error

        return {self.fully_qualified_identifier(): native}

    @classmethod
    async def rpc_handler(cls, request: bytes, context: grpc.aio.ServicerContext) -> bytes:
        """
        Handle the gRPC call to get the list of affected identifiers.

        Args:
          request: The request payload in protobuf ecoding.
          context: The gRPC call context.

        Returns:
          The response payload in protobuf ecoding.
        """

        return cls.encode()

    @typing.override
    @classmethod
    def add_to_feature(cls, feature: "Feature") -> type[typing.Self]:
        super().add_to_feature(feature)

        feature.context.protobuf.register_message(
            name=f"Metadata_{cls.identifier}",
            message={
                "key": Element(
                    identifier=cls.identifier,
                    display_name=cls.display_name,
                    description=cls.description,
                    data_type=cls.data_type,
                )
            },
        )

        feature.context.protobuf.register_service(
            feature.identifier,
            {f"Get_FCPAffectedByMetadata_{cls.identifier}": grpc.unary_unary_rpc_method_handler(cls.rpc_handler)},
            package=feature.rpc_package,
        )

        return cls

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        flag = False
        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            pos = reader.cursor
            tag = reader.read_uint32()

            if tag >> 3 == 1:
                reader.expect_type(tag, WireType.LEN)

                try:
                    if issubclass(cls.data_type, List) or (
                        issubclass(cls.data_type, Constrained) and issubclass(cls.data_type.data_type, List)
                    ):
                        if not flag:
                            values = cls.data_type.decode(reader.buffer[pos:end])
                            message.value = values
                            flag = True

                        reader.skip_type(tag & 7)
                    else:
                        message.value = message.value.decode(reader, reader.read_uint32())
                except DecodeError as error:
                    msg = f"Invalid field '{cls.identifier}' in message '{cls.__name__}': {error.message}"
                    raise DecodeError(msg, error.offset, [cls.identifier, *error.path]) from None

            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    @classmethod
    def encode(cls, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        for affected_call in cls.affects:
            String(affected_call).encode(writer, number or 1)

        if number:
            writer.ldelim()

        return writer.finish()



import decimal
import io
import textwrap

import typing_extensions as typing

from sila import datetime
from sila.datetime.date import _format_offset


class Serializer:
    ctx: typing.ClassVar[decimal.Context] = decimal.Context(prec=20)

    def __init__(self, remove_whitespace: bool = False) -> None:
        self.indentation = 0
        self.buffer = io.StringIO()
        self.remove_whitespace = remove_whitespace

    @classmethod
    def float_to_str(cls, value: float) -> str:
        """
        Convert float to exact string representation, i.e. prevent
        scientific notation produced by float repr.

        Args:
          value: The float value to convert to string.

        Returns:
          The string representation.
        """

        return format(cls.ctx.create_decimal(repr(value)).normalize(), "f")

    @classmethod
    def serialize(cls, handler: typing.Callable[["Serializer"], None], remove_whitespace: bool = False) -> str:
        """
        Serialize an object using the given parsers into an XML string.

        Args:
          handler: The handler to serialize the root element
          remove_whitespace: Whether to omit whitespaces in the output.

        Returns:
          The string representation of the XML data.
        """

        serializer = Serializer(remove_whitespace)
        handler(serializer)

        return serializer.result()

    def result(self) -> str:
        """
        Return the result of the serializer.

        Returns:
          The serialized xml.
        """

        return self.buffer.getvalue()

    def start_element(self, name: str) -> typing.Self:
        """
        Write the start of an element with the given name.

        Args:
          name: the raw XML 1.0 name of the element type.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        self.write(f"<{name}>")
        self.indent()

        return self

    def end_element(self, name: str) -> typing.Self:
        """
        Write the end of an element with the given name.

        Args:
          name: the raw XML 1.0 name of the element type.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        self.dedent()
        self.write(f"</{name}>")

        return self

    def write_str(self, element: str, value: str, width: int = 88) -> typing.Self:
        """
        Write a string value surrounded by the given element.

        Args:
          element: The start and end element surrounding the string.
          value: The string value to write into the xml.
          width: The maximum length of characters per line. If the string
             exceeds this width, it is rendered in several lines.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        content = f"<{element}>"

        if len(value) > width:
            content += "\n"
            content += "\n".join(
                "\n".join(
                    textwrap.wrap(
                        line,
                        width,
                        initial_indent="  ",
                        subsequent_indent="  ",
                        break_long_words=False,
                        replace_whitespace=False,
                    )
                )
                for line in value.splitlines()
            )
            content += "\n"
        else:
            content += value

        content += f"</{element}>"

        return self.write(content)

    def write_bool(self, element: str, value: bool) -> typing.Self:
        """
        Write a boolean value surrounded by the given element.

        Args:
          element: The start and end element surrounding the boolean.
          value: The boolean value to write into the xml.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        return self.write(f"<{element}>{'Yes' if value else 'No'}</{element}>")

    def write_int(self, element: str, value: int) -> typing.Self:
        """
        Write an integer value surrounded by the given element.

        Args:
          element: The start and end element surrounding the integer.
          value: The integer value to write into the xml.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        return self.write(f"<{element}>{int(value)}</{element}>")

    def write_float(self, element: str, value: float) -> typing.Self:
        """
        Write a float value surrounded by the given element.

        Args:
          element: The start and end element surrounding the float.
          value: The float value to write into the xml.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        return self.write(f"<{element}>{self.float_to_str(value)}</{element}>")

    def write_date(self, element: str, value: datetime.date) -> typing.Self:
        """
        Write a date value surrounded by the given element.

        Args:
          element: The start and end element surrounding the date.
          value: The date value to write into the xml.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        isoformat = "{:04d}-{:02d}-{:02d}".format(value.year, value.month, value.day)

        if (tzinfo := getattr(value, "tzinfo", None)) and isinstance(tzinfo, datetime.tzinfo):
            isoformat += _format_offset(offset) if (offset := tzinfo.utcoffset(None)) else "Z"

        return self.write(f"<{element}>{isoformat}</{element}>")

    def write_time(self, element: str, value: datetime.time) -> typing.Self:
        """
        Write a time value surrounded by the given element.

        Args:
          element: The start and end element surrounding the time.
          value: The time value to write into the xml.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        isoformat = "{:02d}:{:02d}:{:02d}.{:03d}".format(
            value.hour, value.minute, value.second, value.microsecond // 1000
        )
        if (tzinfo := getattr(value, "tzinfo", None)) and isinstance(tzinfo, datetime.tzinfo):
            isoformat += _format_offset(offset) if (offset := tzinfo.utcoffset(None)) else "Z"

        return self.write(f"<{element}>{isoformat}</{element}>")

    def write_datetime(self, element: str, value: datetime.datetime) -> typing.Self:
        """
        Write a datetime value surrounded by the given element.

        Args:
          element: The start and end element surrounding the datetime.
          value: The datetime value to write into the xml.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        isoformat = "{:04d}-{:02d}-{:02d}".format(value.year, value.month, value.day)

        isoformat += "T"

        isoformat += "{:02d}:{:02d}:{:02d}.{:03d}".format(
            value.hour, value.minute, value.second, value.microsecond // 1000
        )

        if (tzinfo := getattr(value, "tzinfo", None)) and isinstance(tzinfo, datetime.tzinfo):
            isoformat += _format_offset(offset) if (offset := tzinfo.utcoffset(None)) else "Z"

        return self.write(f"<{element}>{isoformat}</{element}>")

    def write(self, value: str) -> typing.Self:
        """
        Writes the value as a new line with the current indentation.

        Args:
          value: The value to write into the xml.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        buffer = textwrap.indent(
            value + ("\n" if not self.remove_whitespace else ""),
            "  " * (self.indentation if not self.remove_whitespace else 0),
        )

        self.buffer.write(buffer)

        return self

    def indent(self, level: int = 1) -> typing.Self:
        """
        Indent the upcoming lines.

        Args:
          level: By how many tabs to indent.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        self.indentation += level

        return self

    def dedent(self, level: int = 1) -> typing.Self:
        """
        Dedent the upcoming lines.

        Args:
          level: By how many tabs to dedent.

        Returns:
          The Serializer instance, allowing for method chaining.
        """

        self.indentation -= level

        return self



import re

import typing_extensions as typing

from .feature_identifier import FeatureIdentifier, identifier


class DataTypeIdentifier(FeatureIdentifier):
    """
    Uniquely identifies a custom data type among all potentially existing data types.

    Attributes:
      data_type: Unique custom data type identifier.

    Examples:
      >>> identifier = FeatureIdentifier("org.silastandard/core/ErrorRecoveryService/v1/DataType/ContinuationOption")
    """

    _pattern: typing.ClassVar[str] = f"{FeatureIdentifier._pattern}/DataType/(?P<data_type>{identifier})"
    _regex: typing.ClassVar[re.Pattern] = re.compile(_pattern, re.IGNORECASE)

    @typing.override
    @classmethod
    def create(cls, originator: str, category: str, feature: str, version: int, data_type: str) -> typing.Self:
        """
        Create a new custom data type identifier from the given components.

        Args:
          originator: The organization who created and owns the feature.
          category: Group features into domains of application.
          feature: Unique feature identifier.
          version: The major feature version.
          data_type: Unique custom data type identifier.

        Returns:
          The created custom data type identifier.
        """

        identifier = super().__new__(cls, f"{originator}/{category}/{feature}/v{version}/DataType/{data_type}")
        identifier._data = {
            "originator": originator,
            "category": category,
            "feature": feature,
            "version": version,
            "data_type": data_type,
        }

        return identifier

    @property
    def data_type(self) -> str:
        """Unique custom data type identifier."""

        return self._data["data_type"]



import functools
import re

import typing_extensions as typing

from .feature_identifier import FeatureIdentifier, identifier


class CommandIdentifier(FeatureIdentifier):
    """
    Uniquely identifies a command among all potentially existing commands.

    Attributes:
      command: Unique command identifier.

    Examples:
      >>> identifier = FeatureIdentifier("org.silastandard/core/SiLAService/v1/Command/GetFeatureDefinition")
    """

    _pattern: typing.ClassVar[str] = f"{FeatureIdentifier._pattern}/Command/(?P<command>{identifier})"
    _regex: typing.ClassVar[re.Pattern] = re.compile(_pattern, re.IGNORECASE)

    @typing.override
    @classmethod
    def create(cls, originator: str, category: str, feature: str, version: int, command: str) -> typing.Self:
        """
        Create a new command identifier from the given components.

        Args:
          originator: The organization who created and owns the feature.
          category: Group features into domains of application.
          feature: Unique feature identifier.
          version: The major feature version.
          command: Unique command identifier.

        Returns:
          The created command identifier.
        """

        identifier = super().__new__(cls, f"{originator}/{category}/{feature}/v{version}/Command/{command}")
        identifier._data = {
            "originator": originator,
            "category": category,
            "feature": feature,
            "version": version,
            "command": command,
        }

        return identifier

    @property
    def command(self) -> str:
        """Unique command identifier."""

        return self._data["command"]

    @functools.cached_property
    def command_identifier(self) -> "CommandIdentifier":
        """The fully qualified command identifier."""

        return CommandIdentifier.create(self.originator, self.category, self.feature, self.version, self.command)



import logging
from importlib.metadata import version

from . import server
from .framework import (
    AllowedTypes,
    Any,
    BasicType,
    Binary,
    BinaryDownloadFailed,
    BinaryTransfer,
    BinaryTransferError,
    BinaryTransferErrorType,
    BinaryTransferHandler,
    BinaryUploadFailed,
    Boolean,
    CancelRequest,
    ChannelOptions,
    Characters,
    ClientMessage,
    CloudMetadata,
    Command,
    CommandConfirmation,
    CommandConfirmationResponse,
    CommandExecution,
    CommandExecutionInfo,
    CommandExecutionNotAccepted,
    CommandExecutionNotFinished,
    CommandExecutionRequest,
    CommandExecutionResponse,
    CommandExecutionStatus,
    CommandExecutionUUID,
    CommandIdentifier,
    CommandParameter,
    CommandResponseRequest,
    Constrained,
    Constraint,
    ContentType,
    ContentTypeParameter,
    Context,
    ContextProxy,
    ConversionError,
    Convertable,
    CreateBinaryRequest,
    CreateBinaryResponse,
    CreateBinaryUploadRequest,
    Custom,
    DataType,
    DataTypeIdentifier,
    Date,
    DecodeError,
    DefinedExecutionError,
    DeleteBinaryRequest,
    DeleteBinaryResponse,
    Deserializer,
    DownloadChunkRequest,
    DownloadChunkResponse,
    Duration,
    Element,
    ElementCount,
    EncodeError,
    EndDocument,
    EndElement,
    ErrorIdentifier,
    Feature,
    FeatureIdentifier,
    FrameworkError,
    FrameworkErrorType,
    FullyQualifiedIdentifier,
    GetBinaryInfoRequest,
    GetBinaryInfoResponse,
    Handler,
    Integer,
    IntermediateResponseIdentifier,
    InvalidBinaryTransferUUID,
    InvalidCommandExecutionUUID,
    InvalidMetadata,
    Length,
    List,
    MaximalElementCount,
    MaximalExclusive,
    MaximalInclusive,
    MaximalLength,
    Message,
    Metadata,
    MetadataIdentifier,
    MetadataRequest,
    MetadataResponse,
    MinimalElementCount,
    MinimalExclusive,
    MinimalInclusive,
    MinimalLength,
    Native,
    NoMetadataAllowed,
    ObservableCommand,
    ObservableCommandResponse,
    ObservableProperty,
    ParameterIdentifier,
    ParseError,
    Pattern,
    Property,
    PropertyIdentifier,
    PropertyRequest,
    PropertyResponse,
    Protobuf,
    Reader,
    Real,
    ResponseIdentifier,
    Schema,
    SchemaType,
    Serializable,
    Serializer,
    Server,
    ServerMessage,
    Set,
    SiLAConnectionError,
    SiLAError,
    SIUnit,
    StartElement,
    Stream,
    String,
    Structure,
    Time,
    Timestamp,
    Timezone,
    Token,
    UndefinedExecutionError,
    Unit,
    UnitComponent,
    UnobservableCommand,
    UnobservableCommandResponse,
    UnobservableProperty,
    UploadChunkRequest,
    UploadChunkResponse,
    ValidationError,
    VersionLevel,
    Void,
    WireType,
    Writer,
    check_display_name,
    check_identifier,
    check_uuid,
    check_version,
    create_channel,
    create_server,
)

logging.getLogger(__name__).addHandler(logging.NullHandler())

__version__ = version("unitelabs_sila")

__all__ = [
    "AllowedTypes",
    "Any",
    "BasicType",
    "Binary",
    "BinaryDownloadFailed",
    "BinaryTransfer",
    "BinaryTransferError",
    "BinaryTransferErrorType",
    "BinaryTransferHandler",
    "BinaryUploadFailed",
    "Boolean",
    "CancelRequest",
    "ChannelOptions",
    "Characters",
    "ClientMessage",
    "CloudMetadata",
    "Command",
    "CommandConfirmation",
    "CommandConfirmationResponse",
    "CommandExecution",
    "CommandExecutionInfo",
    "CommandExecutionNotAccepted",
    "CommandExecutionNotFinished",
    "CommandExecutionRequest",
    "CommandExecutionResponse",
    "CommandExecutionStatus",
    "CommandExecutionUUID",
    "CommandIdentifier",
    "CommandParameter",
    "CommandResponseRequest",
    "Constrained",
    "Constraint",
    "ContentType",
    "ContentTypeParameter",
    "Context",
    "ContextProxy",
    "ConversionError",
    "Convertable",
    "CreateBinaryRequest",
    "CreateBinaryResponse",
    "CreateBinaryUploadRequest",
    "Custom",
    "DataType",
    "DataTypeIdentifier",
    "Date",
    "DecodeError",
    "DefinedExecutionError",
    "DeleteBinaryRequest",
    "DeleteBinaryResponse",
    "Deserializer",
    "DownloadChunkRequest",
    "DownloadChunkResponse",
    "Duration",
    "Element",
    "ElementCount",
    "EncodeError",
    "EndDocument",
    "EndElement",
    "ErrorIdentifier",
    "Feature",
    "FeatureIdentifier",
    "FrameworkError",
    "FrameworkErrorType",
    "FullyQualifiedIdentifier",
    "GetBinaryInfoRequest",
    "GetBinaryInfoResponse",
    "Handler",
    "Integer",
    "IntermediateResponseIdentifier",
    "InvalidBinaryTransferUUID",
    "InvalidCommandExecutionUUID",
    "InvalidMetadata",
    "Length",
    "List",
    "MaximalElementCount",
    "MaximalExclusive",
    "MaximalInclusive",
    "MaximalLength",
    "Message",
    "Metadata",
    "MetadataIdentifier",
    "MetadataRequest",
    "MetadataResponse",
    "MinimalElementCount",
    "MinimalExclusive",
    "MinimalInclusive",
    "MinimalLength",
    "Native",
    "NoMetadataAllowed",
    "ObservableCommand",
    "ObservableCommandResponse",
    "ObservableProperty",
    "ParameterIdentifier",
    "ParseError",
    "Pattern",
    "Property",
    "PropertyIdentifier",
    "PropertyRequest",
    "PropertyResponse",
    "Protobuf",
    "Reader",
    "Real",
    "ResponseIdentifier",
    "SIUnit",
    "Schema",
    "SchemaType",
    "Serializable",
    "Serializer",
    "Server",
    "ServerMessage",
    "Set",
    "SiLAConnectionError",
    "SiLAError",
    "StartElement",
    "Stream",
    "String",
    "Structure",
    "Time",
    "Timestamp",
    "Timezone",
    "Token",
    "UndefinedExecutionError",
    "Unit",
    "UnitComponent",
    "UnobservableCommand",
    "UnobservableCommandResponse",
    "UnobservableProperty",
    "UploadChunkRequest",
    "UploadChunkResponse",
    "ValidationError",
    "VersionLevel",
    "Void",
    "WireType",
    "Writer",
    "check_display_name",
    "check_identifier",
    "check_uuid",
    "check_version",
    "create_channel",
    "create_server",
    "server",
]



import re


def check_uuid(value: str) -> str:
    """
    Validate whether the given value is a Universally Unique IDentifier according to RFC 4122.

    Args:
      value: The input value to validate.

    Returns:
      The input value, allowing for method chaining.

    Raises:
      ValueError: If the provided value does not follow the expected
        schema.
    """

    if not isinstance(value, str):
        msg = f"UUID must be a string of unicode characters, received '{value}'."
        raise ValueError(msg)

    if not re.fullmatch(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", value):
        msg = f"UUID may only contain letters and digits, received '{value}'."
        raise ValueError(msg)

    return value



import dataclasses

import typing_extensions as typing

from ..protobuf import Message, Reader, WireType, Writer


@dataclasses.dataclass
class MetadataRequest(Message):
    """
    Message the get calls affected by the given metadata.

    Attributes:
      fully_qualified_metadata_id: The fully qualified identifier of
        the metadata to get the affected calls for.
    """

    fully_qualified_metadata_id: str = ""

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message.fully_qualified_metadata_id = reader.read_string()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.fully_qualified_metadata_id:
            writer.write_uint32(10).write_string(self.fully_qualified_metadata_id)

        if number:
            writer.ldelim()

        return writer.finish()



import datetime
import ipaddress

from .discovery import Discovery

try:
    from cryptography import x509
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    INSTALLED = True
except ImportError:
    INSTALLED = False


def generate_certificate(server_uuid: str, host: str) -> tuple[bytes, bytes]:
    """
    Generate a self-signed certificate according to the SiLA 2 standard.

    In short, the Common Name is set to "SiLA2" and the custom OID
    `1.3.6.1.4.1.58583` is set to the provided server uuid. In
    addition to localhost, the certificate uses the grpc server's ip
    address as an alternative name.

    Args:
      server_uuid: The server's uuid used for the custom SiLA OID.
      host: The host on which the grpc server is bound to.

    Returns:
      A tuple containing first the private key as PEM-encoded bytes
      and second the certificate as PEM-encoded bytes.
    """

    if not INSTALLED:
        msg = "Cannot import 'cryptography'"
        raise ImportError(msg)

    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    private_key_content = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    subject = issuer = x509.Name([x509.NameAttribute(x509.OID_COMMON_NAME, "SiLA2")])
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
        .add_extension(
            x509.SubjectAlternativeName(
                [
                    x509.DNSName("localhost"),
                    x509.IPAddress(ipaddress.IPv4Address(Discovery.find_ip_address(host))),
                ]
            ),
            critical=False,
        )
        .add_extension(
            x509.UnrecognizedExtension(x509.ObjectIdentifier("1.3.6.1.4.1.58583"), str(server_uuid).encode("ascii")),
            critical=False,
        )
        .sign(private_key, hashes.SHA256())
    )
    cert_content = cert.public_bytes(serialization.Encoding.PEM)

    return (private_key_content, cert_content)



import dataclasses
import enum

import typing_extensions as typing

from ..data_types import Duration, Real
from ..protobuf import Message, Reader, WireType, Writer


class CommandExecutionStatus(enum.IntEnum):
    """Provides details about the execution status of a command."""

    WAITING = 0
    """The command is waiting for its execution."""

    RUNNING = 1
    """The command is currently executing."""

    FINISHED_SUCCESSFULLY = 2
    """The command finished successfully."""

    FINISHED_WITH_ERROR = 3
    """The command finished with an error."""


@dataclasses.dataclass
class CommandExecutionInfo(Message):
    """Provides information about the current status of a Command being executed."""

    Status: typing.ClassVar[type[CommandExecutionStatus]] = CommandExecutionStatus

    status: CommandExecutionStatus = CommandExecutionStatus.WAITING
    """The current status of the execution of a command."""

    progress: typing.Optional[Real] = None
    """The estimated progress of a command execution in percent (0 - 100%)."""

    remaining_time: typing.Optional[Duration] = None
    """The estimated remaining execution time of the command."""

    updated_lifetime: typing.Optional[Duration] = None
    """The duration during which a command execution UUID is valid."""

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                message.status = CommandExecutionStatus(reader.read_int32())
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message.progress = Real.decode(reader, reader.read_uint32())
            elif field_number == 3:
                reader.expect_type(tag, WireType.LEN)
                message.remaining_time = Duration.decode(reader, reader.read_uint32())
            elif field_number == 4:
                reader.expect_type(tag, WireType.LEN)
                message.updated_lifetime = Duration.decode(reader, reader.read_uint32())
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.status:
            writer.write_uint32(8).write_int32(self.status)
        if self.progress:
            self.progress.encode(writer, 2)
        if self.remaining_time:
            self.remaining_time.encode(writer, 3)
        if self.updated_lifetime:
            self.updated_lifetime.encode(writer, 4)

        if number:
            writer.ldelim()

        return writer.finish()



import collections.abc
import dataclasses
import functools

import typing_extensions as typing

from ..common import Handler
from ..data_types.any import Any
from ..fdl import Serializable
from ..identifiers import PropertyIdentifier

if typing.TYPE_CHECKING:
    from ..common import Feature
    from ..data_types import DataType
    from ..fdl import Deserializer, Serializer


@dataclasses.dataclass
class Property(Handler, Serializable):
    """Describes certain aspects of a server that do not require an action on the server."""

    observable: bool = False
    """Whether the property returns an observable stream or just a single value."""

    data_type: type["DataType"] = Any
    """The SiLA data type of the property."""

    @functools.cached_property
    @typing.override
    def fully_qualified_identifier(self) -> PropertyIdentifier:
        """Uniquely identifies the property."""

        return PropertyIdentifier.create(**super().fully_qualified_identifier._data, property=self.identifier)

    @typing.override
    def add_to_feature(self, feature: "Feature") -> typing.Self:
        super().add_to_feature(feature)

        feature.properties[self.identifier] = self

        return self

    @typing.override
    def serialize(self, serializer: "Serializer") -> None:
        serializer.start_element("Property")
        serializer.write_str("Identifier", self.identifier)
        serializer.write_str("DisplayName", self.display_name)
        serializer.write_str("Description", self.description)
        serializer.write_bool("Observable", self.observable)
        self.data_type.serialize(serializer)
        if self.errors:
            serializer.start_element("DefinedExecutionErrors")
            for Error in self.errors.values():
                serializer.write_str("Identifier", Error.identifier)
            serializer.end_element("DefinedExecutionErrors")
        serializer.end_element("Property")

    @typing.override
    def deserialize(self, deserializer: "Deserializer") -> collections.abc.Generator[None, typing.Any, typing.Self]: ...



from .command import Command
from .command_confirmation import CommandConfirmation
from .command_execution import CommandExecution
from .command_execution_info import CommandExecutionInfo, CommandExecutionStatus
from .command_execution_uuid import CommandExecutionUUID
from .observable_command import ObservableCommand
from .unobservable_command import UnobservableCommand

__all__ = [
    "Command",
    "CommandConfirmation",
    "CommandExecution",
    "CommandExecutionInfo",
    "CommandExecutionStatus",
    "CommandExecutionUUID",
    "ObservableCommand",
    "UnobservableCommand",
]



from .sila_error import SiLAError


class SiLAConnectionError(SiLAError):
    """
    Error that occurs in the underlying communication infrastructure.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """



import builtins
import collections.abc
import dataclasses
import enum

import typing_extensions as typing

from ..data_types.binary import Binary
from ..data_types.string import String
from ..fdl import Characters, Deserializer, EndElement, Serializer, StartElement
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", str, bytes)


class SchemaType(str, enum.Enum):
    """
    Enumeration of schema types used in the Schema constraint.

    This enumeration defines the types of schemas that can be used
    for validation. It currently supports XML and JSON.
    """

    XML = "Xml"
    """Represents the XML schema type."""

    JSON = "Json"
    """Represents the JSON schema type."""


@dataclasses.dataclass
class Schema(Constraint[T]):
    """
    A constraint that enforces the structure of a value to match a
    specific schema, either XML or JSON.

    This class allows validation of a string or bytes object against
    a specific schema type (XML or JSON). The schema can either be
    provided via a URL or inline content.

    Args:
      type: The schema type, either 'Xml' or 'Json'.
      url: An optional URL pointing to the schema.
      inline: Optional inline content representing the schema.
    """

    Type: typing.ClassVar[builtins.type[SchemaType]] = SchemaType
    """
    Enumeration of schema types used in the Schema constraint.
    """

    type: typing.Literal["Xml", "Json"]
    """
    The schema type, either 'Xml' or 'Json'.
    """

    url: typing.Optional[str] = None
    """
    An optional URL pointing to the schema.
    """

    inline: typing.Optional[str] = None
    """
    Optional inline content representing the schema.
    """

    def __post_init__(self):
        if self.url is None and self.inline is None:
            raise ValueError("Either 'url' or 'inline' must be provided.")

        if self.url is not None and self.inline is not None:
            raise ValueError("'url' and 'inline' cannot both be provided.")

        self.type = self.type.value if isinstance(self.type, SchemaType) else self.type

    async def validate(self, value: T) -> bool:
        if not isinstance(value, (str, bytes)):
            raise TypeError(f"Expected value of type 'str' or 'bytes', received '{type(value).__name__}'.")

        # TODO: Load xml or json schema to validate whether `value` follows that schema

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.start_element("Schema")
        serializer.write_str("Type", self.type)
        if self.url is not None:
            serializer.write_str("Url", self.url)
        if self.inline is not None:
            serializer.write_str("Inline", self.inline)
        serializer.end_element("Schema")

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[builtins.type["BasicType"], builtins.type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (String, Binary)):
            raise ValueError(
                f"Expected constraint's data type to be 'String' or 'Binary', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="Schema")

        # Type
        yield from deserializer.read_start_element(name="Type")
        type = yield from deserializer.read_str()
        try:
            type = SchemaType(type).value
        except ValueError:
            raise ValueError(f"Expected a valid 'Type' value, received '{type}'.") from None
        yield from deserializer.read_end_element(name="Type")

        url: typing.Optional[str] = None
        inline: typing.Optional[str] = None

        token = yield
        if isinstance(token, StartElement):
            if token.name == "Url":
                # Url
                url = yield from deserializer.read_str()
                yield from deserializer.read_end_element(name="Url")
            elif token.name == "Inline":
                # Inline
                inline = yield from deserializer.read_str()
                yield from deserializer.read_end_element(name="Inline")
            else:
                raise ValueError(
                    f"Expected start element with name 'Url' or 'Inline', received start element with name '{token.name}'."
                )
        elif isinstance(token, Characters):
            raise ValueError(
                f"Expected start element with name 'Url' or 'Inline', received characters '{token.value}'."
            )
        elif isinstance(token, EndElement):
            raise ValueError(
                f"Expected start element with name 'Url' or 'Inline', received end element with name '{token.name}'."
            )

        yield from deserializer.read_end_element(name="Schema")

        return cls(type, url=url, inline=inline)



import typing_extensions as typing


@typing.runtime_checkable
class BinaryTransferHandler(typing.Protocol):
    """Handle the transfer of large binaries."""

    async def get_binary(self, binary_transfer_uuid: str) -> bytes:
        """
        Retrieve a large binary by its identifier.

        Args:
          binary_transfer_uuid: A unique identifier (UUID) for the binary
            transfer session from which the large binary was retrieved.

        Returns:
          The actual binary data retrieved from the transfer session.
        """
        ...

    async def set_binary(self, value: bytes) -> str:
        """
        Dispatch a large binary through a binary transfer session.

        Args:
          value: The actual binary data being dispatched through the
            binary transfer session.

        Returns:
          A unique identifier (UUID) for the binary transfer session to
          which the data was dispatched.
        """
        ...



import collections.abc
import dataclasses

import typing_extensions as typing

from sila import datetime

from ..data_types.date import Date
from ..data_types.integer import Integer
from ..data_types.real import Real
from ..data_types.time import Time
from ..data_types.timestamp import Timestamp
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", int, float, datetime.date, datetime.time, datetime.datetime)


@dataclasses.dataclass
class MaximalInclusive(Constraint[T]):
    """
    A constraint that enforces an upper inclusive bound on a value.

    This constraint ensures that a given value does not exceed a
    specified maximum inclusive bound. It can be applied to various
    types, including Integer, Real, Date, Time, and Timestamp. The
    constraint raises errors if the input value exceeds the defined
    upper bound or is of an incorrect type.

    Args:
      value: The upper inclusive limit for the value. The value must
        be of the same type as this attribute and must not exceed
        this limit.
    """

    value: T
    """
    The upper inclusive limit for the value. The value must be of
    the same type as this attribute and must not exceed this
    limit.
    """

    async def validate(self, value: T) -> bool:
        if not isinstance(value, type(self.value)):
            raise TypeError(f"Expected value of type '{type(self.value).__name__}', received '{type(value).__name__}'.")

        if not value <= self.value:
            raise ValueError(
                f"Value '{value}' must be less than or equal to the maximal inclusive limit of '{self.value}'."
            )

        return True

    def serialize(self, serializer: Serializer) -> None:
        if isinstance(self.value, int):
            serializer.write_int("MaximalInclusive", self.value)
        elif isinstance(self.value, float):
            serializer.write_float("MaximalInclusive", self.value)
        elif isinstance(self.value, datetime.date):
            serializer.write_date("MaximalInclusive", self.value)
        elif isinstance(self.value, datetime.time):
            serializer.write_time("MaximalInclusive", self.value)
        elif isinstance(self.value, datetime.datetime):
            serializer.write_datetime("MaximalInclusive", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (Integer, Real, Date, Time, Timestamp)):
            raise ValueError(
                f"Expected constraint's data type to be 'Integer', 'Real', 'Date', 'Time' or 'Timestamp', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="MaximalInclusive")
        if issubclass(data_type, Integer):
            value = yield from deserializer.read_int()
        elif issubclass(data_type, Real):
            value = yield from deserializer.read_float()
        elif issubclass(data_type, Date):
            value = yield from deserializer.read_date()
        elif issubclass(data_type, Time):
            value = yield from deserializer.read_time()
        elif issubclass(data_type, Timestamp):
            value = yield from deserializer.read_datetime()
        yield from deserializer.read_end_element(name="MaximalInclusive")

        return cls(typing.cast(T, value))



import grpc.aio
import typing_extensions as typing

from .channel_options import ChannelOptions


async def create_channel(
    target: str,
    /,
    tls: bool = True,
    root_certificates: typing.Optional[bytes] = None,
    certificate_chain: typing.Optional[bytes] = None,
    private_key: typing.Optional[bytes] = None,
    options: typing.Optional[ChannelOptions] = None,
    **kwargs,
) -> grpc.aio.Channel:
    """
    Create a channel to the target.

    Args:
      target: The target address to connect to.
      tls: Whether to run a secure/TLS channel or a plaintext channel
        (i.e. no TLS), defaults to run with TLS encryption.
      root_certificates: The PEM-encoded root certificates as a byte
        string, or None to retrieve them from a default location
        chosen by gRPC runtime.
      certificate_chain: The PEM-encoded certificate chain as a byte
        string to use or None if no certificate chain should be used.
      private_key: The PEM-encoded private key as a byte string, or
        None if no private key should be used.
      options: Additional options for the underlying gRPC connection.

    Returns:
      The grpc channel connected to cloud client endpoint.
    """

    options = ChannelOptions(options or {})

    if tls:
        credentials = grpc.ssl_channel_credentials(
            root_certificates=root_certificates,
            private_key=private_key,
            certificate_chain=certificate_chain,
        )
        channel = grpc.aio.secure_channel(target, credentials=credentials, options=list(options.items()))
    else:
        channel = grpc.aio.insecure_channel(target, options=list(options.items()))

    return channel



import dataclasses
import datetime

import typing_extensions as typing

from ..fdl import Deserializer, ParseError, Serializer
from ..protobuf import DecodeError, Reader, WireType, Writer
from .binary import Binary
from .boolean import Boolean
from .convertable import Native
from .data_type import BasicType, DataType
from .date import Date
from .integer import Integer
from .real import Real
from .string import String
from .time import Time
from .timestamp import Timestamp
from .void import Void

if typing.TYPE_CHECKING:
    from ..common import Context
    from .constrained import Constrained
    from .list import List
    from .structure import Structure


@dataclasses.dataclass
class Any(BasicType[Native]):
    """Represents information that can be of any SiLA data type, except for a custom data type."""

    value: typing.Union[BasicType, "List", "Structure", "Constrained"] = dataclasses.field(default_factory=Void)

    @property
    def schema(self) -> str:
        """The xml representation of the data type definition."""
        serializer = Serializer(remove_whitespace=True)
        self.value.serialize(serializer)

        return serializer.buffer.getvalue()

    @classmethod
    def from_schema(cls, schema: str, payload: bytes = b"") -> typing.Self:
        """
        Create a new `Any` instance from a given schema and payload.

        Args:
          schema: The xml representation of the data type definition.
          payload: The data type's binary protobuf value.

        Returns:
          The newly create `Any` data type instance.
        """

        deserializer = Deserializer()

        data_type: type[DataType] = deserializer.deserialize(schema, DataType.deserialize)

        from .constrained import Constrained
        from .list import List
        from .structure import Structure

        assert issubclass(data_type, (BasicType, List, Structure, Constrained))

        return cls(value=data_type.decode(payload))

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[Native] = None, /) -> typing.Self:
        from .element import Element
        from .list import List
        from .structure import Structure

        data_type: typing.Union[BasicType, "List", "Structure"] = Void()
        if value is None:
            data_type = await Void.from_native(context, value)
        elif isinstance(value, str):
            data_type = await String.from_native(context, value)
        elif isinstance(value, bytes):
            data_type = await Binary.from_native(context, value)
        elif isinstance(value, bool):
            data_type = await Boolean.from_native(context, value)
        elif isinstance(value, int):
            data_type = await Integer.from_native(context, value)
        elif isinstance(value, float):
            data_type = await Real.from_native(context, value)
        elif isinstance(value, datetime.datetime):
            data_type = await Timestamp.from_native(context, value)
        elif isinstance(value, datetime.date):
            data_type = await Date.from_native(context, value)
        elif isinstance(value, datetime.time):
            data_type = await Time.from_native(context, value)
        elif isinstance(value, list):
            item_type = Void
            items: list[typing.Union["BasicType", "Structure", "Constrained"]] = []
            for child_value in value:
                item_data_type = (await Any.from_native(context, child_value)).value

                if isinstance(item_data_type, List):
                    msg = "List may not contain other lists."
                    raise ValueError(msg)

                if items and type(item_data_type).__name__ is not item_type.__name__:
                    msg = "Only same type lists are allowed."
                    raise ValueError(msg)

                item_type = type(item_data_type)
                items.append(item_data_type)

            data_type = List.create(item_type)(items)
        elif isinstance(value, dict):
            elements: dict[str, Element] = {}
            values: dict[str, DataType] = {}
            for key, child_value in value.items():
                item_data_type = (await Any.from_native(context, child_value)).value

                values[key] = item_data_type
                elements[key] = Element(identifier=key, display_name=key, data_type=type(item_data_type))

            data_type = Structure.create(elements)(values)

        return await cls(value=data_type).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> Native:
        return await self.value.to_native(context)

    @typing.override
    async def validate(self) -> typing.Self:
        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        end = reader.length if length is None else reader.cursor + length

        offset = 0
        schema: str = ""
        payload: bytes = b""

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                schema = reader.read_string()
                offset = reader.cursor - len(schema)
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                payload = reader.read_bytes()
            else:
                reader.skip_type(tag & 7)

        try:
            return cls.from_schema(schema, payload)
        except ParseError as error:
            raise DecodeError(error.message, offset=offset + error.column) from error

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.schema:
            writer.write_uint32(10).write_string(self.schema)
        writer.write_uint32(18).write_bytes(self.value.encode())

        if number:
            writer.ldelim()

        return writer.finish()
